import asyncio
from datetime import date, datetime, timedelta
from io import BytesIO
from logging import getLogger
from typing import (
    Dict,
    List,
    Literal,
    Optional,
    Set,
    TypeVar,
    Union,
)

import csp
import orjson
import pyarrow
import pyarrow.json
import uvloop
from csp import ts
from fastapi import APIRouter, WebSocket
from perspective import Client, Server, Table
from perspective.handlers.starlette import PerspectiveStarletteHandler
from pydantic import Field, PrivateAttr
from starlette.websockets import WebSocketDisconnect
from typing_extensions import TypeAliasType

from csp_gateway.server import ChannelSelection, GatewayChannels, GatewayModule
from csp_gateway.server.web import GatewayWebApp, get_default_responses
from csp_gateway.utils import PickleableQueue, get_args, get_origin, get_thread

__all__ = (
    "psp_schema_to_arrow_schema",
    "create_pyarrow_table",
    "MountPerspectiveTables",
)

T = TypeVar("T")

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

log = getLogger(__name__)

_PSP_ARROW_MAP = {
    int: pyarrow.int64(),
    float: pyarrow.float64(),
    bool: pyarrow.bool_(),
    str: pyarrow.string(),
    date: pyarrow.string(),
    datetime: pyarrow.timestamp("us", tz="UTC"),
}


def psp_schema_to_arrow_schema(psp_schema):
    return pyarrow.schema([(k, _PSP_ARROW_MAP[v]) for k, v in psp_schema.items()])


def perspective_thread(client: Client) -> None:
    # Create event loop for perspective callbacks
    psp_loop = asyncio.new_event_loop()

    # Attach to manager
    client.set_loop_callback(psp_loop.call_soon_threadsafe)

    # Run the perspective callback loop
    psp_loop.run_forever()


def create_pyarrow_table(key_name, data, arrow_schema, date_conversion_set):
    new_data = []
    for item in data:
        flattened_items = item.psp_flatten()
        if key_name:
            for f in flattened_items:
                # TODO better
                f["basket-key"] = key_name
        new_data.extend(flattened_items)

    # convert to arrow
    json = b"\n".join(orjson.dumps(_) for _ in new_data)
    b = BytesIO()
    b.write(json)
    b.seek(0)
    table = pyarrow.json.read_json(
        b,
        parse_options=pyarrow.json.ParseOptions(explicit_schema=arrow_schema, unexpected_field_behavior="ignore"),
    )
    if date_conversion_set:
        table = pyarrow.Table.from_arrays(
            [table.column(name).cast(pyarrow.date32()) if name in date_conversion_set else table.column(name) for name in table.column_names],
            names=table.column_names,
        )
    return table


def pull_data_thread(queue: PickleableQueue, table_insts, arrow_schema_insts, arrow_schema_date_conversions):
    while True:
        try:
            for (table_name, key_name), timeserieses in queue.get().items():
                table = create_pyarrow_table(
                    key_name,
                    timeserieses,
                    arrow_schema_insts[table_name],
                    arrow_schema_date_conversions[table_name],
                )
                stream = pyarrow.BufferOutputStream()
                writer = pyarrow.RecordBatchStreamWriter(stream, table.schema)
                writer.write_table(table)
                writer.close()
                table_insts[table_name].update(stream.getvalue().to_pybytes())
        except KeyboardInterrupt:
            raise
        except Exception:
            log.exception("Error processing perspective")


ExcludedColumns = TypeAliasType("ExcludedColumns", "Union[Set[str], Dict[str, Union[bool, ExcludedColumns]]]")


class MountPerspectiveTables(GatewayModule):
    requires: Optional[ChannelSelection] = []

    tables: ChannelSelection = Field(default_factory=ChannelSelection)
    _unused_tables: Optional[List[str]] = PrivateAttr(default_factory=list)

    limits: Dict[str, int] = Field(
        description="Dict mapping table name to [perspective limit](https://perspective-dev.github.io/guide/explanation/table/options.html)",
        default={},
    )
    default_limit: Optional[int] = Field(None, description="Default limit for all tables, i.e. 1000")
    indexes: Dict[str, Optional[str]] = Field(
        description="Dict mapping table name to [perspective index](https://perspective-dev.github.io/guide/explanation/table/options.html)",
        default={},
    )
    default_index: Optional[str] = Field(None, description="Default index field for all tables, i.e. 'id'")
    architectures: Dict[str, Literal["server", "client-server"]] = Field(
        description="Dict mapping table name to [perspective data architecture](https://perspective-dev.github.io/guide/explanation/architecture.html), default is client-server",
        default={},
    )
    default_architecture: Literal["server", "client-server"] = Field(
        "client-server",
        description="Default architecture for all tables, i.e. 'client-server'",
    )

    layouts: Dict[str, str] = Field(default={})
    default_layout: Optional[str] = Field(
        None,
        description="Default layout to use for all tables if no specific layout is provided.",
    )

    update_interval: timedelta = Field(default=timedelta(seconds=2))
    perspective_field: str = Field(
        None,
        description="Optional field on the channels which has an instance of perspective.Server to use, "
        "such that it can also be used by other GatewayModules with custom table preparation logic.",
    )
    layouts_field: str = Field(
        None,
        description="Optional field on the channels which has a dictionary of layouts to use, "
        "such that it can also be used by other GatewayModules with custom table preparation logic.",
    )
    excluded_table_columns: Dict[str, ExcludedColumns] = Field(
        default={},
        description=(
            "Dictionary from table name to columns (which are attributes on a GatewayStruct) to exclude from perspective. "
            "The columns to exclude can be specified as either be a set of column names or as a dictionary. If specified "
            "as a dictionary, the dictionary is a mapping from attribute name to sub-attributes to exclude. This is defined "
            "recursively so it can be used to exclude fields that are structs of structs."
        ),
    )

    _route: str = "/perspective"
    _server: Server = PrivateAttr(default=None)
    _client: Client = PrivateAttr(default=None)

    _layouts: Dict[str, str] = PrivateAttr(default={})
    _schema_insts: Dict[str, Dict] = PrivateAttr(default_factory=dict)
    _arrow_schema_insts: Dict[str, Dict] = PrivateAttr(default_factory=dict)
    _arrow_schema_date_conversions: Dict[str, Set[str]] = PrivateAttr(default_factory=dict)
    _table_insts: Dict[str, Table] = PrivateAttr(default={})
    _queue: PickleableQueue = PrivateAttr(default_factory=PickleableQueue)

    def _connect_all_tables(self, channels: GatewayChannels) -> None:
        for field in self.tables.select_from(channels):
            edge = channels.get_channel(field)
            excluded_columns = self.excluded_table_columns.get(field, None)
            if isinstance(edge, dict):
                if not edge:
                    raise ValueError(f"No keys defined for dict basket channel {field}.")

                to_flatten = []
                for subfield in edge.keys():
                    to_flatten.append(channels.get_channel(field, subfield))

                # save perspective schema
                schema = channels.get_channel(field, subfield).tstype.typ.psp_schema(excluded_columns)

                # save pyarrow schema
                self.add_table(
                    field,
                    schema,
                    limit=self.limits.get(field, self.default_limit),
                    index=self.indexes.get(field, self.default_index),
                )

                if hasattr(subfield, "name"):
                    self.push_to_perspective(csp.flatten(to_flatten), field, subfield.name)
                else:
                    self.push_to_perspective(csp.flatten(to_flatten), field, subfield)

            else:
                ts_type = channels.get_channel(field).tstype.typ
                if get_origin(ts_type) is list:
                    schema = get_args(ts_type)[0].psp_schema(excluded_columns)
                    edge = csp.unroll(channels.get_channel(field))
                else:
                    schema = ts_type.psp_schema(excluded_columns)
                    edge = channels.get_channel(field)

                self.add_table(
                    field,
                    schema,
                    limit=self.limits.get(field, self.default_limit),
                    index=self.indexes.get(field, self.default_index),
                )
                self.push_to_perspective(
                    edge,
                    field,
                )
            self._schema_insts[field] = schema
            self._arrow_schema_insts[field] = psp_schema_to_arrow_schema(schema)
            self._arrow_schema_date_conversions[field] = set()
            # annoying workaround for pyarrow reading dates
            for k, v in self._schema_insts[field].items():
                if v is date:
                    self._arrow_schema_date_conversions[field].add(k)

    def get_schema_from_field(self, channels: GatewayChannels, field: str):
        edge = channels.get_channel(field)

        if isinstance(edge, dict):
            a_subfield = list(edge.keys())[0]
            edge = channels.get_channel(field, a_subfield)

        ts_type = edge.tstype.typ

        # if its a list of structs
        if get_origin(ts_type) is list:
            struct_type = get_args(ts_type)[0]
        else:
            struct_type = ts_type

        if not hasattr(struct_type, "psp_schema"):
            raise Exception(f"Type has no conversion to perspective: {struct_type}")

        excluded_columns = self.excluded_table_columns.get(field, None)
        return struct_type.psp_schema(excluded_columns)

    def add_table(self, field: str, schema, limit: int = None, index: str = None):
        self._table_insts[field] = self._client.table(schema, limit, index, name=field)

    def connect(self, channels: GatewayChannels) -> None:
        if self.perspective_field:
            self._server = getattr(channels, self.perspective_field)
        else:
            self._server = Server()
        self._client = self._server.new_local_client()
        self._layouts = self.layouts.copy()
        if self.layouts_field:
            self._layouts.update(getattr(channels, self.layouts_field))
        self._connect_all_tables(channels)

        # Run perspective on background daemon threads
        self.run_perspective()

    @csp.node
    def push_to_perspective(  # type: ignore[no-untyped-def]
        self,
        timeseries: ts["T"],
        table_name: str,
        key_name: str = "",
    ):
        with csp.alarms():
            alarm: ts[bool] = csp.alarm(bool)
        with csp.state():
            s_buffer = {}

        with csp.start():
            csp.schedule_alarm(alarm, self.update_interval, True)

        if csp.ticked(timeseries):
            if (table_name, key_name) not in s_buffer:
                s_buffer[(table_name, key_name)] = []
            s_buffer[(table_name, key_name)].append(timeseries)

        if csp.ticked(alarm):
            if len(s_buffer) > 0:
                self._queue.put(s_buffer)
                s_buffer = {}
            csp.schedule_alarm(alarm, self.update_interval, True)

    def _get_tables(self) -> Dict[str, Dict[str, str]]:
        all_tables = {table_name: None for table_name in self._client.get_hosted_table_names() if table_name not in self._unused_tables}
        for table_name in all_tables:
            table = self._client.open_table(table_name)
            schema = table.schema()
            all_tables[table_name] = {col: schema[col] for col in table.columns()}
        return all_tables

    def _get_table_sizes(self) -> Dict[str, int]:
        all_tables = {table_name: None for table_name in self._client.get_hosted_table_names()}
        for table_name in all_tables:
            table = self._client.open_table(table_name)
            all_tables[table_name] = table.size()
        return all_tables

    def _check_unused_tables(self, channels: GatewayChannels) -> None:
        for table, key in channels._null_ts:
            # NOTE: if only a single key is not bound, we consider the table
            # in-use since we flatten the keys
            # TODO: if all keys are unused, we could consider the table unused
            if key is not None or table in self._unused_tables:
                continue
            self._unused_tables.append(table)

    def rest(self, app: GatewayWebApp) -> None:
        # NOTE: We use an internal API here
        self._check_unused_tables(object.__getattribute__(app.gateway, "channels"))

        async def websocket_handler(websocket: WebSocket) -> None:
            handler = PerspectiveStarletteHandler(perspective_server=self._server, websocket=websocket)
            try:
                await handler.run()
            except WebSocketDisconnect:
                # ignore
                ...

        # Get API Router
        api_router: APIRouter = app.get_router("api")

        # Mount the perspective websocket handler
        api_router.add_api_websocket_route(self._route, websocket_handler)

        # add route to fetch table names
        @api_router.get(
            "{}/{}".format(self._route, "tables"),
            responses=get_default_responses(),
            response_model=Dict[str, Dict[str, str]],
            tags=["Utility"],
        )
        async def get_perspective_table_names():
            """
            This endpoint exposes the served [perspective](https://github.com/perspective-dev/perspective) table names and schemas.

            It can be used for `perspective`-based dashboards to build interactive visualization,
            such as the one provided as an example for the `csp-gateway` project.
            Depending on your server's configuration, this might be available at [`/`](/)
            """
            return self._get_tables()

        # add route to fetch layouts
        @api_router.get(
            "{}/{}".format(self._route, "layouts"),
            responses=get_default_responses(),
            response_model=Dict[str, str],
            tags=["Utility"],
        )
        async def get_perspective_layouts():
            """
            This endpoint exposes saved [perspective](https://github.com/perspective-dev/perspective) workspace layouts.
            These layouts can be configured server side to be provided to all clients.
            """
            return self._layouts

        # add route to fetch layouts
        @api_router.get(
            "{}/{}".format(self._route, "meta"),
            responses=get_default_responses(),
            response_model=Dict[str, Union[None, int, str, Dict[str, Union[str, int]], Dict[str, Dict[str, str]], List[str]]],
            tags=["Utility"],
        )
        async def get_perspective_meta():
            """
            This endpoint exposes perspective meta information, including limits,
            indexes, and architecture.
            """
            return {
                "limits": self.limits,
                "default_limit": self.default_limit,
                "indexes": self.indexes,
                "default_index": self.default_index,
                "architectures": self.architectures,
                "default_architecture": self.default_architecture,
                "layouts": self.layouts,
                "default_layout": self.default_layout,
                "tables": self._get_tables(),
                "unused_tables": self._unused_tables,
                "table_sizes": self._get_table_sizes(),
            }

    def run_perspective(self):
        """Launch the perspective threads"""
        psp_load_thread = get_thread(
            target=pull_data_thread,
            args=(
                self._queue,
                self._table_insts,
                self._arrow_schema_insts,
                self._arrow_schema_date_conversions,
            ),
        )
        psp_load_thread.start()



import pickle

from csp_gateway.utils.picklable_queue import PickleableQueue


def test_queue_pickling():
    queue = PickleableQueue()
    unpickled_queue = pickle.loads(pickle.dumps(queue))

    assert queue.queue == unpickled_queue.queue
    assert queue.maxsize == unpickled_queue.maxsize



import abc
import typing
from datetime import datetime
from typing import Any, Dict, Generic, List, Optional, Set, Type, Union

from ccflow import BaseModel
from pydantic import Field, TypeAdapter, model_validator

from csp_gateway.server.shared import ChannelSelection
from csp_gateway.utils import GatewayStruct

from .channels import ChannelsType

if typing.TYPE_CHECKING:
    from csp_gateway.server import GatewayWebApp


class Module(BaseModel, Generic[ChannelsType], abc.ABC):
    model_config = {"arbitrary_types_allowed": True}

    requires: Optional[ChannelSelection] = None
    disable: bool = False
    block_set_channels_until: Optional[datetime] = Field(
        default=None,
        description="""
        This determines the csp time at which this module can start sending data to channels.
        This value overrides any gateway-level blocks imposed.
        """,
    )

    @abc.abstractmethod
    def connect(self, Channels: ChannelsType) -> None: ...

    def rest(self, app: "GatewayWebApp") -> None: ...

    @abc.abstractmethod
    def shutdown(self) -> None: ...

    def dynamic_keys(self) -> Optional[Dict[str, List[Any]]]: ...

    def dynamic_channels(self) -> Optional[Dict[str, Union[Type[GatewayStruct], Type[List[GatewayStruct]]]]]:
        """
        Channels that this module dynamically adds to the gateway channels when this module is included into the gateway.

        Returns:
            Dictionary keyed by channel name and type of the timeseries of the channel as values.
        """
        ...

    def dynamic_state_channels(self) -> Optional[Set[str]]:
        """
        The set of dynamic channels that have state.
        """
        ...

    # @abc.abstractmethod
    # def subscribe(self):
    #     ...

    def __eq__(self, other):
        # Override equality because occasionally, Modules will contain fields with non-standard equality methods
        # i.e. numpy arrays or csp edges.
        # Without overriding, these types will prevent the modules from being compared with each other
        # which is needed for the dependency resolutions
        return id(self) == id(other)

    # See https://docs.pydantic.dev/latest/concepts/validators/#validation-of-default-values
    @model_validator(mode="before")
    def validate_requires(cls, v):
        requires = v.get("requires", cls.model_fields["requires"].default)
        v["requires"] = TypeAdapter(ChannelSelection).validate_python(requires)
        return v



import typing
from datetime import datetime

from csp import Enum, Struct

from csp_gateway import Filter, FilterCondition, Query, State, StateType, disable_duckdb_state, enable_duckdb_state, modify_buffer_threshold

_STATES = [StateType.DUCKDB, StateType.DEFAULT]
_BUFFER_THRESHOLD = 10000
_BUFFER_SIZES = [int(_BUFFER_THRESHOLD / 10), _BUFFER_THRESHOLD - 1, _BUFFER_THRESHOLD, _BUFFER_THRESHOLD + 1, _BUFFER_THRESHOLD * 10]
_FILTER = Filter(attr="a", by=FilterCondition(value=int(_BUFFER_THRESHOLD / 10), where="<="))
_QUERY = Query(filters=[_FILTER, _FILTER])


class MyEnum(Enum):
    A = 1
    B = 2
    C = 3


class CspSubStruct(Struct):
    suba: int = 1
    subb: str = "b"
    subc: datetime = datetime.now()
    subd: str
    sube: typing.Dict[str, str]
    subf: MyEnum = MyEnum.A


class NonCspStruct(object):
    a: int = 0
    b: str
    c: datetime = datetime.now()
    d: str = ""
    e: typing.Dict[str, str] = {"A": "hhello", "B": "bbye"}
    f: MyEnum = MyEnum.A
    g: CspSubStruct = CspSubStruct()

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class CspStruct(Struct):
    a: int = 0
    b: str
    c: datetime = datetime.now()
    d: str = ""
    e: typing.Dict[str, str] = {"A": "hhello", "B": "bbye"}
    f: MyEnum = MyEnum.A
    h: typing.List[int] = [1, 2, 4]
    i: list = [1, 2, 4]
    j: dict = {"a": "b"}
    g: CspSubStruct = CspSubStruct()


class StateInitialize:
    params = _STATES
    param_names = ["StateType"]

    def setup(self, state_typ):
        if state_typ == StateType.DEFAULT:
            disable_duckdb_state()
        elif state_typ == StateType.DUCKDB:
            enable_duckdb_state()
        return None

    def time_create(self, state_typ):
        _ = State[CspStruct](keyby="a")


class StateInsert:
    params = [_STATES, _BUFFER_SIZES]
    param_names = ["StateType", "#Inserts"]
    timeout = 300

    def setup(self, state_typ, _):
        if state_typ == StateType.DEFAULT:
            disable_duckdb_state()
        elif state_typ == StateType.DUCKDB:
            enable_duckdb_state()
        self.s = State[CspStruct]()
        modify_buffer_threshold(_BUFFER_THRESHOLD)

    def time_insert(self, state_typ, threshold):
        for i in range(threshold):
            ts = CspStruct(a=i, b=f"{i}", f=MyEnum.A, h=[i, i, i], j={i: i}, g=CspSubStruct(suba=i + 1, subb=f"{i + 1}"))
            self.s.insert(ts)


class StateInsertKeyBy:
    params = [_STATES, _BUFFER_SIZES]
    param_names = ["StateType", "#Inserts"]
    timeout = 300

    def setup(self, state_typ, _):
        if state_typ == StateType.DEFAULT:
            disable_duckdb_state()
        elif state_typ == StateType.DUCKDB:
            enable_duckdb_state()
        self.s = State[CspStruct](keyby="a")
        modify_buffer_threshold(_BUFFER_THRESHOLD)

    def time_insert(self, state_typ, threshold):
        for i in range(threshold):
            # Have the keyby value repeat
            ts = CspStruct(a=i % 100, b=f"{i}", f=MyEnum.A, h=[i, i, i], j={i: i}, g=CspSubStruct(suba=i + 1, subb=f"{i + 1}"))
            self.s.insert(ts)


class StateFirstQueryAll:
    params = [_STATES, _BUFFER_SIZES]
    param_names = ["StateType", "#Queries"]
    timeout = 300

    def setup(self, state_typ, threshold):
        if state_typ == StateType.DEFAULT:
            disable_duckdb_state()
        elif state_typ == StateType.DUCKDB:
            enable_duckdb_state()
        self.s = State[CspStruct](keyby="a")
        modify_buffer_threshold(_BUFFER_THRESHOLD)
        for i in range(threshold):
            ts = CspStruct(a=i, b=f"{i}", f=MyEnum.A, h=[i, i, i], j={i: i}, g=CspSubStruct(suba=i + 1, subb=f"{i + 1}"))
            self.s.insert(ts)

    def time_query(self, state_typ, _):
        self.s.query()

    time_query.number = 1


class StateMultiQueryAll:
    params = [_STATES, _BUFFER_SIZES]
    param_names = ["StateType", "#Queries"]
    timeout = 300

    def setup(self, state_typ, threshold):
        if state_typ == StateType.DEFAULT:
            disable_duckdb_state()
        elif state_typ == StateType.DUCKDB:
            enable_duckdb_state()
        self.s = State[CspStruct](keyby="a")
        modify_buffer_threshold(_BUFFER_THRESHOLD)
        for i in range(threshold):
            ts = CspStruct(a=i, b=f"{i}", f=MyEnum.A, h=[i, i, i], j={i: i}, g=CspSubStruct(suba=i + 1, subb=f"{i + 1}"))
            self.s.insert(ts)

    def time_query(self, state_typ, _):
        self.s.query()


class StateFirstQueryFilter:
    params = [_STATES, _BUFFER_SIZES]
    param_names = ["StateType", "#Queries"]
    timeout = 300

    def setup(self, state_typ, threshold):
        if state_typ == StateType.DEFAULT:
            disable_duckdb_state()
        elif state_typ == StateType.DUCKDB:
            enable_duckdb_state()
        self.s = State[CspStruct](keyby="a")
        modify_buffer_threshold(_BUFFER_THRESHOLD)
        for i in range(threshold):
            ts = CspStruct(a=i, b=f"{i}", f=MyEnum.A, h=[i, i, i], j={i: i}, g=CspSubStruct(suba=i + 1, subb=f"{i + 1}"))
            self.s.insert(ts)

    def time_query(self, state_typ, _):
        self.s.query(_QUERY)

    time_query.number = 1


class StateMultiQueryFilter:
    params = [_STATES, _BUFFER_SIZES]
    param_names = ["StateType", "#Queries"]
    timeout = 300

    def setup(self, state_typ, threshold):
        if state_typ == StateType.DEFAULT:
            disable_duckdb_state()
        elif state_typ == StateType.DUCKDB:
            enable_duckdb_state()
        self.s = State[CspStruct](keyby="a")
        modify_buffer_threshold(_BUFFER_THRESHOLD)
        for i in range(threshold):
            ts = CspStruct(a=i, b=f"{i}", f=MyEnum.A, h=[i, i, i], j={i: i}, g=CspSubStruct(suba=i + 1, subb=f"{i + 1}"))
            self.s.insert(ts)

    def time_query(self, state_typ, _):
        self.s.query(_QUERY)



import logging
from abc import abstractmethod
from asyncio import (
    AbstractEventLoop,
    Future,
    Queue,
    get_event_loop,
    new_event_loop,
    run_coroutine_threadsafe,
    set_event_loop,
    wrap_future,
)
from copy import deepcopy
from functools import lru_cache
from json import JSONDecodeError
from threading import Thread
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Literal, Optional, Tuple, Union, cast

from httpx import AsyncClient as httpx_AsyncClient, Response, get as GET, post as POST
from jsonref import replace_refs
from nest_asyncio import apply as applyAsyncioNesting
from packaging import version
from pydantic import Field, PrivateAttr, field_validator, model_validator

try:
    from ccflow import BaseModel
except ImportError:
    from pydantic import BaseModel
try:
    from orjson import loads
except ImportError:
    from json import loads

if TYPE_CHECKING:
    from aiohttp import ClientSession
    from pandas import DataFrame as PandasDataFrame
    from polars import DataFrame as PolarsDataFrame

from ..utils import (
    Query,
    ServerRouteNotFoundException,
    ServerRouteNotMountedException,
    ServerUnknownException,
    ServerUnprocessableException,
    get_thread,
)

__all__ = (
    "GatewayClientConfig",
    "ResponseWrapper",
    "ResponseType",
    "BaseGatewayClient",
    "SyncGatewayClientMixin",
    "SyncGatewayClient",
    "AsyncGatewayClientMixin",
    "AsyncGatewayClient",
    "Client",
    "AsyncClient",
    "ClientConfig",
    "GatewayClient",
    "GatewayClientConfiguration",
    "ClientConfiguration",
)

log = logging.getLogger(__name__)

_DEFAULT_TIMEOUT = 10.0
_PANDAS_TYPE_MAPPING = {
    "string": "string",
    "integer": "int64",
    "number": "float64",
    "boolean": "bool",
    "object": "object",
}

_POLARS_TYPE_MAPPING = {
    "string": "String",
    "integer": "Int64",
    "number": "Float64",
    "boolean": "Boolean",
    "object": "Object",
}


def _openapi_to_dataframe_dtypes(schema, path, df_type_mapping):
    res = {}
    for column, properties in schema["properties"].items():
        if "properties" in properties:
            res.update(_openapi_to_dataframe_dtypes(properties, path + column + ".", df_type_mapping))
            continue
        if "anyOf" in properties:
            # This schema value can be one of many types
            new_multi_type = list(filter(lambda d: d.get("type") != "null", properties["anyOf"]))
            if len(new_multi_type) > 1:
                # Multiple non-null types, keep the type as object
                openapi_type = "object"
            else:
                properties = new_multi_type[0]
                # Only 1 non-null type, use that as the type
                if "properties" in properties:
                    res.update(_openapi_to_dataframe_dtypes(properties, path + column + ".", df_type_mapping))
                    continue
                openapi_type = new_multi_type[0].get("type")
        else:
            openapi_type = properties.get("type")
        dataframe_type = df_type_mapping.get(openapi_type, df_type_mapping.get("object"))
        res[path + column] = dataframe_type
    return res


def _normalize_field(field: str) -> str:
    if "/" in field:
        return field.split("/")[0] + "/"
    return field


def _get_schema(spec: Dict[str, Any], path: str) -> Optional[Dict[str, Any]]:
    if (paths := spec.get("paths")) is None:
        return None

    if (path_data := paths.get(path)) is None:
        # could be a lookup for an id, or a key
        base_path = path.rpartition("/")[0]
        if (possible_path := paths.get(base_path + "/{id}")) is not None:
            path_data = possible_path
        elif (possible_path := paths.get(base_path + "/{key}")) is not None:
            path_data = possible_path
        else:
            return None

    response_type = "get" if "get" in path_data else "post"
    try:
        return path_data[response_type]["responses"]["200"]["content"]["application/json"]["schema"]
    except KeyError:
        return None


def _raiseIfNotMounted(foo: Callable) -> Callable:
    group = foo.__name__

    def _wrapped_foo(self: Any, field: str = "", *args: Any, **kwargs: Any) -> Any:
        if not self._initialized:
            self._initialize()
        check_field = _normalize_field(field)

        if check_field and check_field not in self._mounted_apis[group]:
            raise ServerRouteNotMountedException("Route not mounted in group {}: {}".format(group, field))
        return foo(self, field, *args, **kwargs)

    return _wrapped_foo


@lru_cache(maxsize=1)
def _host(config) -> str:
    host = config.host
    if not host.startswith(("http://", "https://")):
        if config.port == 443:
            # Force https
            host = f"https://{host}"
        elif config.port == 80:
            host = f"http://{host}"
        else:
            host = f"{config.protocol}://{host}"
    if host.endswith("/"):
        host = host[:-1]
    if config.port:
        if config.port not in (80, 443):
            host += f":{config.port}"

    return host


class GatewayClientConfig(BaseModel):
    model_config = dict(extra="forbid")  # To raise error on misspelling/misspecification

    protocol: Literal["http", "https"] = "http"
    host: str = "localhost"
    port: Optional[int] = Field(default=8000, ge=1, le=65535, description="Port number for the gateway server")
    api_route: str = "/api/v1"
    authenticate: bool = False
    api_key: str = ""
    return_raw_json: bool = Field(
        True, description="Determines whether REST request responses should be returned as raw json data, or as a ResponseWrapper object."
    )

    @model_validator(mode="after")
    def validate_config(self):
        if self.port is not None and self.port < 1:
            raise ValueError("Port must be a positive integer")
        if self.api_key and not self.authenticate:
            raise ValueError("API key must be provided if authentication is enabled")
        if self.host.startswith("http"):
            # Switch protocol to host
            protocol, host = self.host.split("://")
            self.__dict__["protocol"] = protocol
            self.__dict__["host"] = host
        return self

    def __hash__(self):
        return hash(self.model_dump_json())


class ResponseWrapper(BaseModel):
    json_data: Any
    openapi_schema: Optional[Dict[str, Any]] = None

    @field_validator("openapi_schema", mode="after")
    def validate_openapi_schema(cls, v):
        if v is not None:
            # we have a list of items, so we get the properties for those
            if (schema := v.get("items")) is not None:
                return schema
        return v

    def is_empty(self):
        return bool(self.json_data)

    def as_json(self):
        return self.json_data

    def as_pandas_df(self) -> "PandasDataFrame":
        try:
            import pandas as pd
        except ImportError:
            log.exception("Must have pandas installed to use `as_pandas_df`")
            raise
        if self.openapi_schema is None:
            log.warning(f"No matching schema found for {self.json_data}")
            return pd.json_normalize(self.json_data or {})
        if "properties" not in self.openapi_schema:
            # we have a list of non-dicts, return a single column dataframe
            typ = _PANDAS_TYPE_MAPPING.get(self.openapi_schema.get("type"), "object")
            return pd.DataFrame(self.json_data or {}).astype(typ)
        pandas_dtypes = _openapi_to_dataframe_dtypes(self.openapi_schema, "", _PANDAS_TYPE_MAPPING)
        if self.json_data:
            res = pd.json_normalize(self.json_data)
            # we do this to not change the order of columns
            columns = {col: True for col in res.columns}
            for col in pandas_dtypes.keys():
                if col not in columns:
                    columns[col] = True
            # We might have some columns missing data,
            # such as attributes on the structs that are dicts.
            res = res.reindex(columns=columns.keys(), fill_value=None)
            return res.astype(dict(pandas_dtypes.items()))
        # No data
        else:
            return pd.DataFrame({k: pd.Series(dtype=v) for k, v in pandas_dtypes.items()})

    def as_polars_df(self) -> "PolarsDataFrame":
        try:
            import polars as pl
        except ImportError:
            log.exception("Must have polars installed to use `as_polars_df`")
            raise

        if version.parse(pl.__version__) < version.parse("1.0.0"):
            raise ValueError("Polars version must be 1.0.0 or higher to use json_normalize")

        if self.openapi_schema is None:
            log.warning(f"No matching schema found for {self.json_data}")
            return pl.json_normalize(self.json_data or {})
        if "properties" not in self.openapi_schema:
            typ = getattr(pl, _POLARS_TYPE_MAPPING.get(self.openapi_schema.get("type"), "Object"))
            return pl.DataFrame(self.json_data or []).cast(typ)

        polars_dtypes = _openapi_to_dataframe_dtypes(self.openapi_schema, "", {k: getattr(pl, v) for k, v in _POLARS_TYPE_MAPPING.items()})
        return pl.json_normalize(self.json_data, schema=polars_dtypes)


ResponseType = Union[ResponseWrapper, Dict[str, Any]]


def _get_or_new_event_loop() -> AbstractEventLoop:
    try:
        return get_event_loop()
    except RuntimeError:
        log.debug("Attempted to get event loop without one running, creating a new one...")
        # We need a new event loop if this is running on another thread
        set_event_loop(new_event_loop())
        return get_event_loop()


class BaseGatewayClient(BaseModel):
    # server configuration
    config: GatewayClientConfig = Field(default_factory=GatewayClientConfig)

    # openapi configureation
    _initialized: bool = PrivateAttr(default=False)
    _openapi_spec: Dict[Any, Any] = PrivateAttr(default=None)
    _mounted_apis: Dict[str, set] = PrivateAttr(
        default={
            "controls": set(),
            "last": set(),
            "lookup": set(),
            "next": set(),
            "send": set(),
            "state": set(),
        }
    )

    # stream configuration
    _initialized_streaming: Future = PrivateAttr(default=None)
    _request_queue: Queue = PrivateAttr(default=None)
    _event_loop_setup: bool = PrivateAttr(default=False)
    _event_loop: Optional[AbstractEventLoop] = PrivateAttr(default=None)
    _event_loop_thread: Optional[Thread] = PrivateAttr(default=None)

    def __init__(self, config: GatewayClientConfig = None, **kwargs) -> None:
        # Exists for compatibility with positional argument instantiation
        if config is None:
            config = GatewayClientConfig()
        if kwargs:
            config = GatewayClientConfig(**{**config.model_dump(exclude_unset=True), **kwargs})
        super().__init__(config=config)

    @model_validator(mode="after")
    def validate_client(self):
        if self._event_loop is None:
            self._event_loop = _get_or_new_event_loop()

        if self._initialized_streaming is None:
            self._initialized_streaming = Future(loop=self._event_loop)

        if self._request_queue is None:
            self._request_queue = Queue()

        return self

    def _initializeStreaming(self) -> None:
        if not self._initialized_streaming.done():
            # handle streaming / event loop
            if self._event_loop.is_running():
                # loop is already setup, assume in main thread and use as-is
                self._event_loop_setup = True
            else:
                # setup a new event loop in thread
                self._event_loop_thread = get_thread(target=self._event_loop.run_forever)
                self._event_loop_thread.start()

            self._initialized_streaming.set_result(True)

    def _initialize(self) -> None:
        if not self._initialized:
            # grab openapi spec
            self._openapi_spec: Dict[Any, Any] = replace_refs(
                cast(Dict[Any, Any], GET(f"{_host(self.config)}/openapi.json")).json(),
            )

            # collect mounted routes
            for path in self._openapi_spec["paths"]:
                path = path.replace(self.config.api_route, "")
                for group, subroute in (
                    ("controls", "/controls/"),
                    ("last", "/last/"),
                    ("lookup", "/lookup/"),
                    ("next", "/next/"),
                    ("send", "/send/"),
                    ("state", "/state/"),
                ):
                    if "{channel}" in path:
                        # TODO later
                        continue

                    if path.startswith(subroute):
                        key = path.replace(subroute, "")

                        if "{key}" in key:
                            # dict basket
                            key = key.replace("{key}", "")

                        if "/{id}" in key:
                            # lookup route
                            key = key.replace("/{id}", "")

                        self._mounted_apis[group].add(key)

        # mark self as initialized
        self._initialized = True

    def _buildpath(self, route: str) -> str:
        return f"{self.config.api_route}/{route}"

    def _buildroute(self, route: str) -> str:
        url = f"{_host(self.config)}{self._buildpath(route)}"
        if self.config.authenticate:
            return url, {"token": self.config.api_key}
        return url, {}

    def _api_path_and_route(self, route: str) -> str:
        return self.config.api_route + "/" + route

    def _buildroutews(self, route: str) -> str:
        host = _host(self.config)
        if host.startswith("http://"):
            host = host.replace("http://", "ws://")
        elif host.startswith("https://"):
            host = host.replace("https://", "wss://")
        if self.config.authenticate:
            auth = f"?token={self.config.api_key}"
        else:
            auth = ""
        return f"{host}{self.config.api_route}/{route}{auth}"

    def _handle_response(self, resp: Response, route: str) -> ResponseType:
        try:
            resp_json = cast(Dict[str, Any], resp.json())
        except JSONDecodeError as e:
            resp_json = dict(detail=str(e))
        if resp.status_code == 200:
            if self.config.return_raw_json:
                return resp_json
            path = self._buildpath(route=route)
            schema = _get_schema(spec=self._openapi_spec, path=path)
            return ResponseWrapper(json_data=resp_json, openapi_schema=schema)
        elif resp.status_code == 404:
            raise ServerRouteNotFoundException(resp_json.get("detail"))
        elif resp.status_code == 422:
            raise ServerUnprocessableException(resp_json.get("detail"))
        raise ServerUnknownException(f"{resp.status_code}: {resp_json.get('detail')}")

    def _get(
        self,
        route: str,
        params: Dict[str, Any] = None,
        timeout: float = _DEFAULT_TIMEOUT,
    ) -> ResponseType:
        params = params or {}
        resolved_route, extra_params = self._buildroute(route)
        return self._handle_response(GET(resolved_route, params={**params, **extra_params}, timeout=timeout), route=route)

    async def _getasync(
        self,
        route: str,
        params: Dict[str, Any] = None,
        timeout: float = _DEFAULT_TIMEOUT,
    ) -> ResponseType:
        params = params or {}
        resolved_route, extra_params = self._buildroute(route)
        async with httpx_AsyncClient() as client:
            return self._handle_response(await client.get(resolved_route, params={**params, **extra_params}, timeout=timeout), route=route)

    def _post(
        self,
        route: str,
        params: Dict[str, Any] = None,
        data: Dict[str, Any] = None,
        timeout: float = _DEFAULT_TIMEOUT,
    ) -> ResponseType:
        params = params or {}
        resolved_route, extra_params = self._buildroute(route)
        return self._handle_response(POST(resolved_route, params={**params, **extra_params}, json=data, timeout=timeout), route=route)

    async def _postasync(
        self,
        route: str,
        params: Dict[str, Any] = None,
        data: Dict[str, Any] = None,
        timeout: float = _DEFAULT_TIMEOUT,
    ) -> ResponseType:
        params = params or {}
        resolved_route, extra_params = self._buildroute(route)
        async with httpx_AsyncClient() as client:
            return self._handle_response(
                await client.post(resolved_route, params={**params, **extra_params}, json=data, timeout=timeout), route=route
            )

    def _stream(
        self,
        channels: Optional[List[Union[str, Tuple[str, str]]]] = None,
        callback: Callable = None,
    ):
        if callback:
            async_generator = self._streamAsync(channels=channels)
            iterator = async_generator.__aiter__()

            async def wait_for_aio_fut(aio_fut):
                return await aio_fut

            try:
                if self._event_loop.is_running():
                    applyAsyncioNesting(self._event_loop)
                while True:
                    callback(self._event_loop.run_until_complete(iterator.__anext__()))
            except StopAsyncIteration:
                return

    async def _streamAsync(self, channels: Optional[List[Union[str, Tuple[str, str]]]] = None):
        if not self._initialized_streaming.done():
            self._initializeStreaming()

        channels = channels or []

        for channel in channels:
            if isinstance(channel, (list, tuple)) and len(channel) == 2:
                true_channel, key = channel
                await self._subscribe(true_channel, key)
            else:
                await self._subscribe(channel)

        async for data in self._connectAsync():
            yield data

    async def _subscribe(self, channel: str, key: Optional[str] = None):
        if not self._initialized_streaming:
            await self._initialized_streaming

        subscription = dict(action="subscribe", channel=channel)
        if key is not None:
            subscription["key"] = key

        if self._event_loop_setup:
            return await self._request_queue.put(subscription)

        return wrap_future(
            run_coroutine_threadsafe(
                self._request_queue.put(subscription),
                loop=self._event_loop,
            ),
            loop=self._event_loop,
        )

    async def _unsubscribe(self, channel: str, key: Optional[str] = None):
        if not self._initialized_streaming:
            await self._initialized_streaming

        subscription_removal = dict(action="unsubscribe", channel=channel)
        if key is not None:
            subscription_removal["key"] = key

        if self._event_loop_setup:
            return await self._request_queue.put(subscription_removal)

        return wrap_future(
            run_coroutine_threadsafe(
                self._request_queue.put(subscription_removal),
                loop=self._event_loop,
            ),
            loop=self._event_loop,
        )

    async def _publish(self, channel: str, data: Union[Dict[str, Any], List[Any]], key: Optional[str] = None):
        if not self._initialized_streaming:
            await self._initialized_streaming
        send_msg = dict(action="send", channel=channel, data=data)
        if key is not None:
            send_msg["key"] = key
        if self._event_loop_setup:
            return await self._request_queue.put(send_msg)

        return wrap_future(
            run_coroutine_threadsafe(
                self._request_queue.put(send_msg),
                loop=self._event_loop,
            ),
            loop=self._event_loop,
        )

    async def _websocketData(self, ws):
        from aiohttp import WSMsgType

        async for msg in ws:
            if msg.type == WSMsgType.TEXT:
                yield ("out", loads(msg.data))
            elif msg.type == WSMsgType.CLOSED:
                break
            elif msg.type == WSMsgType.ERROR:
                break

    async def _clientRequests(self):
        # infinite generator to deal with client requests
        if self._event_loop_setup:
            while True:
                yield ("in", await self._request_queue.get())
        else:
            while True:
                yield (
                    "in",
                    await wrap_future(
                        run_coroutine_threadsafe(self._request_queue.get(), loop=self._event_loop),
                        loop=self._event_loop,
                    ),
                )

    def _aiohttp_session(self) -> "ClientSession":
        try:
            from aiohttp import ClientSession

            return ClientSession()
        except ImportError:
            log.exception("Must have aiohttp installed to use WebSocket streaming")
            raise

    async def _connectAsync(self):
        from aiostream.stream import merge

        session = self._aiohttp_session()

        route = self._buildroutews("stream")

        async with session.ws_connect(route) as ws:
            # merge async generators
            merged = merge(self._websocketData(ws), self._clientRequests())

            async with merged.stream() as streamer:
                async for direction, data in streamer:
                    if direction == "in":
                        # send to server
                        await ws.send_json(data)
                    elif direction == "out":
                        # yield out to client
                        yield data
                    else:
                        # ignore
                        ...

    # Public Functions

    @property
    def openapi_spec(self):
        self._initialize()
        return deepcopy(self._openapi_spec)

    @abstractmethod
    def controls(self, field: str = "", data: Any = None, timeout: float = _DEFAULT_TIMEOUT) -> ResponseType: ...

    @abstractmethod
    def last(self, field: str = "", timeout: float = _DEFAULT_TIMEOUT) -> ResponseType: ...

    @abstractmethod
    def lookup(self, field: str, id: str, timeout: float = _DEFAULT_TIMEOUT) -> ResponseType: ...

    @abstractmethod
    def next(self, field: str = "", timeout: float = None) -> ResponseType: ...

    @abstractmethod
    def send(self, field: str = "", data: Any = None, timeout: float = _DEFAULT_TIMEOUT) -> ResponseType: ...

    @abstractmethod
    def state(self, field: str = "", timeout: float = _DEFAULT_TIMEOUT, query: Optional[Query] = None) -> ResponseType: ...

    @abstractmethod
    def stream(self, channels: Optional[List[Union[str, Tuple[str, str]]]] = None): ...

    # NOTE: sync version
    # def stream(self, channels: List[str] = None, callback: Callable = None):

    @abstractmethod
    def subscribe(self, field: str = "", key: Optional[str] = None): ...

    @abstractmethod
    def publish(self, field: str, data: Any, key: Optional[str] = None): ...

    @abstractmethod
    def unsubscribe(self, field: str = "", key: Optional[str] = None): ...


class SyncGatewayClientMixin:
    @_raiseIfNotMounted
    def controls(
        self, field: str = "", data: Any = None, timeout: float = _DEFAULT_TIMEOUT, return_raw_json_override: Optional[bool] = None
    ) -> ResponseType:
        if return_raw_json_override is not None:
            old_return_raw_json = self.config.return_raw_json
            self.config.return_raw_json = return_raw_json_override
        if field in ("shutdown",):
            res = self._post("{}/{}".format("controls", field), data=data, timeout=timeout)
        else:
            res = self._get("{}/{}".format("controls", field), timeout=timeout)
        if return_raw_json_override is not None:
            self.config.return_raw_json = old_return_raw_json
        return res

    @_raiseIfNotMounted
    def last(self, field: str = "", timeout: float = _DEFAULT_TIMEOUT, return_raw_json_override: Optional[bool] = None) -> ResponseType:
        if return_raw_json_override is not None:
            old_return_raw_json = self.config.return_raw_json
            self.config.return_raw_json = return_raw_json_override
        res = self._get("{}/{}".format("last", field), timeout=timeout)
        if return_raw_json_override is not None:
            self.config.return_raw_json = old_return_raw_json
        return res

    @_raiseIfNotMounted
    def lookup(self, field: str, id: str, timeout: float = _DEFAULT_TIMEOUT, return_raw_json_override: Optional[bool] = None) -> ResponseType:
        if return_raw_json_override is not None:
            old_return_raw_json = self.config.return_raw_json
            self.config.return_raw_json = return_raw_json_override
        res = self._get("{}/{}/{}".format("lookup", field, id), timeout=timeout)
        if return_raw_json_override is not None:
            self.config.return_raw_json = old_return_raw_json
        return res

    @_raiseIfNotMounted
    def next(self, field: str = "", timeout: float = None, return_raw_json_override: Optional[bool] = None) -> ResponseType:
        if return_raw_json_override is not None:
            old_return_raw_json = self.config.return_raw_json
            self.config.return_raw_json = return_raw_json_override
        res = self._get("{}/{}".format("next", field), timeout=timeout)
        if return_raw_json_override is not None:
            self.config.return_raw_json = old_return_raw_json
        return res

    @_raiseIfNotMounted
    def send(
        self, field: str = "", data: Any = None, timeout: float = _DEFAULT_TIMEOUT, return_raw_json_override: Optional[bool] = None
    ) -> ResponseType:
        if return_raw_json_override is not None:
            old_return_raw_json = self.config.return_raw_json
            self.config.return_raw_json = return_raw_json_override
        res = self._post("{}/{}".format("send", field), data=data, timeout=timeout)
        if return_raw_json_override is not None:
            self.config.return_raw_json = old_return_raw_json
        return res

    @_raiseIfNotMounted
    def state(
        self, field: str = "", timeout: float = _DEFAULT_TIMEOUT, query: Optional[Query] = None, return_raw_json_override: Optional[bool] = None
    ) -> ResponseType:
        if return_raw_json_override is not None:
            old_return_raw_json = self.config.return_raw_json
            self.config.return_raw_json = return_raw_json_override
        params = None if query is None else {"query": query.model_dump_json()}
        res = self._get("{}/{}".format("state", field), timeout=timeout, params=params)
        if return_raw_json_override is not None:
            self.config.return_raw_json = old_return_raw_json
        return res

    def stream(self, channels: Optional[List[Union[str, Tuple[str, str]]]] = None, callback: Callable = None):
        """Stream data from specified channels with optional key filtering for dict baskets.

        Establishes a synchronous streaming connection to receive real-time updates from the specified channels.
        For dict basket channels, you can subscribe to specific keys by providing tuples of (channel, key).

        Args:
            channels: A list of channel names to subscribe to. For dict basket channels,
                    each entry can be either a string (channel name) or a tuple of
                    (channel_name, key) to subscribe only to a specific key in a dict basket.
            callback: A function that will be called with each received message.
        """
        self._stream(channels=channels, callback=callback)

    def publish(self, field: str, data: Union[Dict[str, Any], List[Any]], key: Optional[str] = None):
        """Publish data to a channel or specific key within a dict basket channel.

        This synchronous method sends data to a specific channel. For dict basket channels,
        a key can be specified to send data to that specific key.

        Args:
            field: The channel name to publish to.
            data: The data to publish.
            key: For dict basket channels, the specific key to publish to. If None, publishes to the entire channel.
        """
        self._event_loop.run_until_complete(self._publish(channel=field, data=data, key=key))

    def subscribe(self, field: str = "", key: Optional[str] = None):
        """Subscribe to a channel or specific key within a dict basket channel.

        This synchronous method subscribes to data updates from a specific channel. For dict basket channels,
        a key can be specified to subscribe only to updates for that specific key.

        Args:
            field: The channel name to subscribe to.
            key: For dict basket channels, the specific key to subscribe to. If None, subscribes to the entire channel.
        """
        self._event_loop.run_until_complete(self._subscribe(channel=field, key=key))

    def unsubscribe(self, field: str = "", key: Optional[str] = None):
        """Unsubscribe from a channel or specific key within a dict basket channel.

        This synchronous method unsubscribes from data updates from a specific channel. For dict basket channels,
        a key can be specified to unsubscribe only from updates for that specific key.

        Args:
            field: The channel name to unsubscribe from.
            key: For dict basket channels, the specific key to unsubscribe from. If None, unsubscribes from the entire channel.
        """
        self._event_loop.run_until_complete(self._unsubscribe(channel=field, key=key))


class SyncGatewayClient(SyncGatewayClientMixin, BaseGatewayClient):
    """
    A synchronous (blocking) Gateway client.

    This client will expose synchronous methods to connect to the server's REST endpoints:

        - controls
        - last
        - lookup
        - next
        - send
        - state

    These methods will depend on the configuration of the corresponding GatewayServer.

    Additionally, if the server is configured with WebSocket support, the following syncronous methods will be available:

        - stream
        - publish
        - subscribe
        - unsubscribe
    """

    ...


class AsyncGatewayClientMixin:
    @_raiseIfNotMounted
    async def controls(self, field: str = "", data: Any = None, timeout: float = _DEFAULT_TIMEOUT) -> ResponseType:
        if field in ("shutdown",):
            return await self._postasync("{}/{}".format("controls", field), data=data, timeout=timeout)
        return await self._getasync("{}/{}".format("controls", field), timeout=timeout)

    @_raiseIfNotMounted
    async def last(self, field: str = "", timeout: float = _DEFAULT_TIMEOUT) -> ResponseType:
        return await self._getasync("{}/{}".format("last", field), timeout=timeout)

    @_raiseIfNotMounted
    async def lookup(self, field: str, id: str, timeout: float = _DEFAULT_TIMEOUT) -> ResponseType:
        return await self._getasync("{}/{}/{}".format("lookup", field, id), timeout=timeout)

    @_raiseIfNotMounted
    async def next(self, field: str = "", timeout: float = None) -> ResponseType:
        return await self._getasync("{}/{}".format("next", field), timeout=timeout)

    @_raiseIfNotMounted
    async def send(self, field: str = "", data: Any = None, timeout: float = _DEFAULT_TIMEOUT) -> ResponseType:
        return await self._postasync("{}/{}".format("send", field), data=data, timeout=timeout)

    @_raiseIfNotMounted
    async def state(self, field: str = "", timeout: float = _DEFAULT_TIMEOUT, query: Optional[Query] = None) -> ResponseType:
        params = None if query is None else {"query": query.model_dump_json()}
        return await self._getasync("{}/{}".format("state", field), timeout=timeout, params=params)

    async def stream(self, channels: List[Union[str, Tuple[str, str]]] = None):
        """Stream data from specified channels with optional key filtering for dict baskets.

        Establishes an asynchronous streaming connection to receive real-time updates from the specified channels.
        For dict basket channels, you can subscribe to specific keys by providing tuples of (channel, key).

        Args:
            channels: A list of channel names to subscribe to. For dict basket channels,
                    each entry can be either a string (channel name) or a tuple of
                    (channel_name, key) to subscribe only to a specific key in a dict basket.

        Yields:
            Data messages received from the subscribed channels.
        """
        async for data in self._streamAsync(channels=channels):
            yield data

    async def publish(self, field: str, data: Union[Dict[str, Any], List[Any]], key: Optional[str] = None):
        """Publish data to a channel or specific key within a dict basket channel.

        This asynchronous method sends data to a specific channel. For dict basket channels,
        a key can be specified to send data to that specific key.

        Args:
            field: The channel name to publish to.
            data: The data to publish.
            key: For dict basket channels, the specific key to publish to. If None, publishes to the entire channel.
        """
        await self._publish(channel=field, data=data, key=key)

    async def subscribe(self, field: str, key: Optional[str] = None):
        """Subscribe to a channel or specific key within a dict basket channel.

        This asynchronous method subscribes to data updates from a specific channel. For dict basket channels,
        a key can be specified to subscribe only to updates for that specific key.

        Args:
            field: The channel name to subscribe to.
            key: For dict basket channels, the specific key to subscribe to. If None, subscribes to the entire channel.
        """
        await self._subscribe(channel=field, key=key)

    async def unsubscribe(self, field: str, key: Optional[str] = None):
        """Unsubscribe from a channel or specific key within a dict basket channel.

        This asynchronous method unsubscribes from data updates from a specific channel. For dict basket channels,
        a key can be specified to unsubscribe only from updates for that specific key.

        Args:
            field: The channel name to unsubscribe from.
            key: For dict basket channels, the specific key to unsubscribe from. If None, unsubscribes from the entire channel.
        """
        await self._unsubscribe(channel=field, key=key)


class AsyncGatewayClient(AsyncGatewayClientMixin, BaseGatewayClient):
    """
    An asynchronous Gateway client.

    This client will expose async methods:
    - controls
    - last
    - lookup
    - next
    - send
    - state

    These methods will depend on the configuration of the corresponding GatewayServer.

    Additionally, if the server is configured with WebSocket support, the following asyncronous methods will be available:

        - stream
        - publish
        - subscribe
        - unsubscribe
    """

    ...


Client = SyncGatewayClient
AsyncClient = AsyncGatewayClient
GatewayClient = SyncGatewayClient

ClientConfig = GatewayClientConfig
GatewayClientConfiguration = GatewayClientConfig
ClientConfiguration = GatewayClientConfig



"""
A Demo CSP Gateway application

NOTE: The webserver tests use this code internally to validate rest endpoints,
and those tests in turn ensure that this demo works.
"""

from datetime import date, datetime, timedelta
from logging import INFO, basicConfig
from pathlib import Path
from typing import Annotated, Dict, List

import csp
import numpy as np
from csp import Enum, ts
from csp.typing import Numpy1DArray
from perspective import Server as PerspectiveServer, Table as PerspectiveTable
from pydantic import AfterValidator, Field, field_validator

from csp_gateway import (
    Controls,
    Gateway,
    GatewayChannels,
    GatewayModule,
    GatewaySettings,
    GatewayStructMixins,
    IdType,
    MountAPIKeyMiddleware,
    MountChannelsGraph,
    MountControls,
    MountOutputsFolder,
    MountPerspectiveTables,
    MountRestRoutes,
    MountWebSocketRoutes,
    State,
)
from csp_gateway.server.config import load_gateway

SPEED = timedelta(seconds=1)
SCALE = 10

# Let's walk through a worked example.

# First, we want to define the edges that will be available to our modules.
# In `csp-gateway` dialect, these deferred edges are called `channels`.
# Let's define a few `csp` `Struct` and `Enum` to go with them.

# NOTE: See below for the entry point to run this example

__all__ = (
    "ExampleCspStruct",
    "ExampleData",
    "ExampleEnum",
    "ExampleGatewayChannels",
    "ExampleModule",
    "ExampleModuleCustomTable",
    "ExampleModuleFeedback",
    "ExampleModuleCustomTable",
)


class ExampleCspStruct(csp.Struct):
    z: int = 12


def nonnegative_check(v):
    if v < 0:
        raise ValueError("value must be non-negative.")
    return v


class ExampleDataBase(csp.Struct):
    x: Annotated[int, AfterValidator(nonnegative_check)]
    y: str = ""
    internal_csp_struct: ExampleCspStruct = ExampleCspStruct()
    data: Numpy1DArray[float] = np.array([])
    mapping: Dict[str, int] = {}
    dt: datetime = datetime.today()
    d: date = datetime.today().date()

    @classmethod
    def __get_validator_dict__(cls):
        return {"_validate_example": field_validator("x", mode="after")(nonnegative_check)}


# We can just add in the mixins to make an existing csp.Struct
# csp-gateway compatible, but we also could have defined
# ExampleDataBase to inherit from GatewayStruct directly
class ExampleData(*GatewayStructMixins, ExampleDataBase):
    id: IdType
    timestamp: datetime


class ExampleEnum(Enum):
    A = 1
    B = 2
    C = 3

    def __lt__(self, other):
        # so csp doesn't complain about memoization
        return self.value < other.value


# `MyGatewayChannels` is the collection of lazily-connected `csp` edges that will be provided to all modules in our graph.
# The modules in our graph are `pydantic`-wrapped `csp` modules, that use the `channel` APIs to connect to edges.
# Let's define a simple one.


class ExampleGatewayChannels(GatewayChannels):
    metadata: Dict[str, str] = {"name": "Demo"}

    example: ts[ExampleData] = None
    example_list: ts[List[ExampleData]] = None
    never_ticks: ts[ExampleData] = None

    s_example: ts[State[ExampleData]] = None

    basket: Dict[ExampleEnum, ts[ExampleData]] = None
    str_basket: Dict[str, ts[ExampleData]] = None

    # FIXME
    # basket_list: Dict[ExampleEnum, ts[[ExampleData]]] = None
    # NOTE: this second one is not populated with data so will 404

    # Controls channels for webserver/graph admin
    controls: ts[Controls] = None

    # Custom perspective manager, shared across modules to allow for custom table definitions
    perspective: PerspectiveServer = Field(default_factory=PerspectiveServer)


# A `GatewayModule` has two important features.
# First, it is a typed `pydantic` model, so you can define attributes in the usual `pydantic` way.
# Second, it has a `connect` method that will be provided the `GatewayChannels` instance when the graph is eventually wired together.
# You can use `get_channel` and `set_channel` to read-from and publish-to `csp` edges, respectively.


class ExampleModule(GatewayModule):
    interval: timedelta = SPEED

    # An example module that ticks some data in a struct
    @csp.node
    def subscribe(
        self,
        trigger: ts[bool],
        multiplier: int = 3,
    ) -> ts[ExampleData]:
        with csp.state():
            last_x = 0
        if csp.ticked(trigger):
            last_x += 1
            return ExampleData(
                x=last_x,
                y=str(last_x) * multiplier,
                data=np.random.random((SCALE,)),
                mapping={str(last_x): last_x},
            )

    # An example module that ticks some data in a [struct]
    @csp.node
    def subscribe_list(
        self,
        data: ts[ExampleData],
    ) -> ts[List[ExampleData]]:
        if csp.ticked(data):
            return [data]

    def dynamic_keys(self):
        return {ExampleGatewayChannels.str_basket: ["a", "b", "c"]}

    def connect(self, channels: ExampleGatewayChannels):
        # Create some CSP data streams
        data = self.subscribe(csp.timer(interval=self.interval, value=True))
        data_list = self.subscribe_list(data)

        # Channels set via `set_channel`
        channels.set_channel(ExampleGatewayChannels.example, data)
        channels.set_channel(ExampleGatewayChannels.example_list, data_list)

        # Generic channel for sending data from non-csp sources
        channels.add_send_channel(ExampleGatewayChannels.example)
        channels.add_send_channel(ExampleGatewayChannels.basket, ExampleEnum.A)
        channels.add_send_channel(ExampleGatewayChannels.basket, ExampleEnum.B)
        channels.add_send_channel(ExampleGatewayChannels.basket, ExampleEnum.C)
        channels.add_send_channel(ExampleGatewayChannels.basket)

        # Rudimentary state accumulation via `set_state`
        channels.set_state(ExampleGatewayChannels.example, "id")

        # Create some data streams for dict baskets
        data_a = self.subscribe(
            csp.timer(interval=self.interval, value=True),
            multiplier=ExampleEnum.A.value,
        )
        data_b = self.subscribe(
            csp.timer(interval=self.interval, value=True),
            multiplier=ExampleEnum.B.value,
        )
        data_c = self.subscribe(
            csp.timer(interval=self.interval, value=True),
            multiplier=ExampleEnum.C.value,
        )

        # Dict basket channels can be set as a whole, or via individual keys
        channels.set_channel(ExampleGatewayChannels.basket, data_a, ExampleEnum.A)
        channels.set_channel(ExampleGatewayChannels.basket, data_b, ExampleEnum.B)
        channels.set_channel(ExampleGatewayChannels.basket, data_c, ExampleEnum.C)

        channels.set_channel(ExampleGatewayChannels.str_basket, data_a, "a")
        channels.set_channel(ExampleGatewayChannels.str_basket, data_b, "b")
        channels.set_channel(ExampleGatewayChannels.str_basket, data_c, "c")


class ExampleModuleFeedback(GatewayModule):
    interval: timedelta = SPEED

    # An example module that ticks some data in a struct
    @csp.node
    def subscribe(
        self,
        data: ts[ExampleData],
    ) -> ts[ExampleData]:
        if csp.ticked(data):
            if data.x % 2 == 0:
                return ExampleData(
                    x=data.x + 1,
                    y=data.y,
                )

    def connect(self, channels: ExampleGatewayChannels):
        # Channels set via `set_channel`
        data = channels.get_channel(ExampleGatewayChannels.example)
        fb_data = self.subscribe(data)
        channels.set_channel(ExampleGatewayChannels.example, fb_data)

        # Dict basket channels can be set as a whole, or via individual keys
        channels.set_channel(ExampleGatewayChannels.basket, fb_data, ExampleEnum.A)


class ExampleModuleCustomTable(GatewayModule):
    """This shows how you can customize the perspective tables."""

    table_name: str = "my_custom_table"

    def connect(self, channels: ExampleGatewayChannels):
        perspective_client = channels.perspective.new_local_client()
        my_table = perspective_client.table(dict(timestamp=datetime, x=int, y=str), limit=None, index="y", name=self.table_name)

        example = channels.get_channel(ExampleGatewayChannels.example)
        example_list = csp.unroll(channels.get_channel(ExampleGatewayChannels.example_list))
        # Merge multiple channels into a single stream for display
        # One could also do joins or any other custom combination/transformation of channels
        merged_data = csp.merge(example, example_list)
        self.push_to_perspective(merged_data, my_table)

    @csp.node
    def push_to_perspective(  # type: ignore[no-untyped-def]
        self,
        timeseries: ts[ExampleData],
        table: PerspectiveTable,
    ):
        with csp.alarms():
            alarm: ts[bool] = csp.alarm(bool)
        with csp.state():
            s_buffer = []

        with csp.start():
            csp.schedule_alarm(alarm, timedelta(seconds=0.5), True)

        if csp.ticked(timeseries):
            s_buffer.append(
                {
                    "timestamp": timeseries.timestamp.timestamp(),
                    "x": timeseries.x,
                    "y": timeseries.y,
                }
            )

        if csp.ticked(alarm):
            if len(s_buffer) > 0:
                table.update(s_buffer)
                s_buffer.clear()
            csp.schedule_alarm(alarm, timedelta(seconds=0.5), True)


if __name__ == "__main__":
    # csp-gateway is configured as a hydra application, but it can also
    # be instantiated directly as we do so here:

    # Setting authentication
    settings = GatewaySettings(API_KEY="12345", AUTHENTICATE=False)

    # instantiate gateway
    gateway = Gateway(
        settings=settings,
        modules=[
            ExampleModule(),
            ExampleModuleFeedback(),
            ExampleModuleCustomTable(),
            MountChannelsGraph(),
            MountControls(),
            MountOutputsFolder(),
            MountPerspectiveTables(
                perspective_field="perspective",
                layouts={
                    "Server Defined Layout": '{"sizes":[1],"detail":{"main":{"type":"split-area","orientation":"vertical","children":[{"type":"split-area","orientation":"horizontal","children":[{"type":"tab-area","widgets":["EXAMPLE_LIST_GENERATED_4"],"currentIndex":0},{"type":"tab-area","widgets":["PERSPECTIVE_GENERATED_ID_1"],"currentIndex":0}],"sizes":[0.3,0.7]},{"type":"split-area","orientation":"horizontal","children":[{"type":"tab-area","widgets":["EXAMPLE_GENERATED_3"],"currentIndex":0},{"type":"tab-area","widgets":["PERSPECTIVE_GENERATED_ID_0"],"currentIndex":0}],"sizes":[0.3,0.7]}],"sizes":[0.5,0.5]}},"viewers":{"EXAMPLE_LIST_GENERATED_4":{"version":"3.3.4","plugin":"Datagrid","plugin_config":{"columns":{},"edit_mode":"READ_ONLY","scroll_lock":false},"columns_config":{},"title":"example_list","group_by":[],"split_by":[],"columns":["timestamp","x","y","data","mapping","dt","d","internal_csp_struct.z"],"filter":[],"sort":[["timestamp","desc"]],"expressions":{},"aggregates":{},"table":"example_list","settings":false},"PERSPECTIVE_GENERATED_ID_1":{"version":"3.3.4","plugin":"X Bar","plugin_config":{},"columns_config":{},"title":"example_list (*)","group_by":["x"],"split_by":[],"columns":["y"],"filter":[],"sort":[["x","asc"]],"expressions":{},"aggregates":{"y":"median"},"table":"example_list","settings":false},"EXAMPLE_GENERATED_3":{"version":"3.3.4","plugin":"Datagrid","plugin_config":{"columns":{},"edit_mode":"READ_ONLY","scroll_lock":false},"columns_config":{},"title":"example","group_by":[],"split_by":[],"columns":["timestamp","x","y","data","mapping","dt","d","internal_csp_struct.z"],"filter":[],"sort":[["timestamp","desc"]],"expressions":{},"aggregates":{},"table":"example","settings":false},"PERSPECTIVE_GENERATED_ID_0":{"version":"3.3.4","plugin":"Treemap","plugin_config":{},"columns_config":{},"title":"example (*)","group_by":["x"],"split_by":[],"columns":["y","x",null],"filter":[],"sort":[["timestamp","desc"]],"expressions":{},"aggregates":{},"table":"example","settings":false}}}'  # noqa: E501
                },
                update_interval=timedelta(seconds=1),
            ),
            MountRestRoutes(force_mount_all=True),
            MountWebSocketRoutes(),
            # For authentication
            MountAPIKeyMiddleware(),
        ],
        channels=ExampleGatewayChannels(),
    )

    # To run, we could run our object directly:
    # gateway.start(rest=True, ui=True)

    # But instead, lets run the same code via hydra
    # We can use our own custom config, in config/demo.yaml
    # which inherits from csp-gateway's example config.
    #
    # With hydra, we can easily construct hierarchichal,
    # extensible configurations for all our modules!
    gateway = load_gateway(
        overrides=["+config=omnibus"],
        config_dir=Path(__file__).parent,
    )

    # Set our log level to info. If using hydra,
    # we have even more configuration options at our disposal
    basicConfig(level=INFO)

    # Start the gateway
    gateway.start(rest=True, ui=True)

    # You can also run this directly via cli
    # > pip install csp-gateway
    # > csp-gateway-start --config-dir=csp_gateway/server/demo +config=omnibus



from datetime import date, datetime, timedelta, timezone
from typing import Dict, List, Optional
from unittest.mock import MagicMock

import csp
import numpy as np
import polars as pl
import pytest
from csp import Enum, ts
from csp.typing import Numpy1DArray
from perspective import Server
from pydantic import Field

from csp_gateway import (
    Gateway,
    GatewayChannels,
    GatewayModule,
    GatewayStruct,
    MountPerspectiveTables,
    create_pyarrow_table,
    psp_schema_to_arrow_schema,
)
from csp_gateway.testing.harness import GatewayTestHarness
from csp_gateway.utils.struct import GatewayStructMixins


class MyTestEnum(Enum):
    a = Enum.auto()
    b = Enum.auto()
    c = Enum.auto()


class MyTestSubStruct(GatewayStruct):
    y: List[int]
    x: int = 1
    z: float


class MyTestStruct(GatewayStruct):
    sub: MyTestSubStruct
    z: MyTestEnum
    ncsp_sub: object


class MyTestOptionalSubStruct(GatewayStruct):
    sub_a: Optional[int]


class MyTestOptionalStruct(GatewayStruct):
    i: int
    i_o: Optional[int]
    ob: object
    ob_op: Optional[object]
    l_i: [int]
    lt_i: List[int]
    lt_i_op: Optional[List[int]]
    s: MyTestSubStruct
    s_op: Optional[MyTestSubStruct]
    e: MyTestEnum
    e_op: Optional[MyTestEnum]
    sub: MyTestOptionalSubStruct
    sub_op: Optional[MyTestOptionalSubStruct]


class MyPyArrowStruct(GatewayStruct):
    d: date


class GWC(GatewayChannels):
    test_channel: ts[MyTestStruct] = None
    list_channel: ts[List[MyTestStruct]] = None
    dict_channel: Dict[str, ts[MyTestStruct]] = None
    dict_enum_channel: Dict[MyTestEnum, ts[MyTestStruct]] = None
    index_channel: ts[MyTestStruct] = None
    limit_channel: ts[MyTestStruct] = None
    exclude_channel: ts[MyTestStruct] = None

    my_perspective: Server = Field(default_factory=Server)


class GWCUnused(GatewayChannels):
    a: ts[MyTestStruct] = None
    b: ts[MyTestStruct] = None
    c: ts[[MyTestStruct]] = None
    d: ts[[MyTestStruct]] = None
    e: Dict[MyTestEnum, ts[MyTestStruct]] = None
    f: Dict[MyTestEnum, ts[MyTestStruct]] = None


def test_inherited_container_annotation_schema():
    # Base declares container field
    from csp import Struct

    class BaseStruct(Struct):
        arr: List[int]

    # Child has no own annotations for arr; relies on inherited annotation
    class ChildStruct(*GatewayStructMixins, BaseStruct):
        # explicit id/timestamp, like typical mixin usage
        id: str
        timestamp: datetime

    schema = ChildStruct.psp_schema()
    # Ensure container element type is preserved as int, not str
    assert schema["arr"] is int


def test_inherited_ndarray_annotation_schema():
    from csp import Struct

    # Base declares numpy array field with element type
    class BaseStruct(Struct):
        arr_nd: Numpy1DArray[float] = np.array([1.0]).view(Numpy1DArray)

    # Child inherits without redeclaring annotation
    class ChildStruct(*GatewayStructMixins, BaseStruct):
        id: str
        timestamp: datetime

    schema = ChildStruct.psp_schema()
    # Ensure ndarray element type is preserved as float, not str
    assert schema["arr_nd"] is float


def test_inherited_container_on_gatewaystruct():
    # Parent defines a container field on GatewayStruct
    class ParentGS(GatewayStruct):
        arr: List[int]

    # Child does not redeclare annotations
    class ChildGS(ParentGS):
        pass

    schema = ChildGS.psp_schema()
    assert schema["arr"] is int


def test_inherited_ndarray_on_gatewaystruct():
    class ParentGSArr(GatewayStruct):
        arr_nd: Numpy1DArray[float] = np.array([1.0]).view(Numpy1DArray)

    class ChildGSArr(ParentGSArr):
        pass

    schema = ChildGSArr.psp_schema()
    assert schema["arr_nd"] is float


def test_recursive_perspective_schema():
    schema = MyTestStruct.psp_schema()
    assert isinstance(schema, dict)
    assert schema == {
        "id": str,
        "timestamp": datetime,
        "sub.id": str,
        "sub.y": int,
        "sub.x": int,
        "sub.z": float,
        "sub.timestamp": datetime,
        "z": str,
    }

    schema = MyTestOptionalStruct.psp_schema()
    assert isinstance(schema, dict)
    assert schema == {
        "id": str,
        "timestamp": datetime,
        "i": int,
        "i_o": int,
        "l_i": int,
        "lt_i": int,
        "lt_i_op": int,
        "s.id": str,
        "s.y": int,
        "s.x": int,
        "s.z": float,
        "s.timestamp": datetime,
        "s_op.id": str,
        "s_op.y": int,
        "s_op.x": int,
        "s_op.z": float,
        "s_op.timestamp": datetime,
        "e": str,
        "e_op": str,
        "sub.id": str,
        "sub.sub_a": int,
        "sub.timestamp": datetime,
        "sub_op.id": str,
        "sub_op.sub_a": int,
        "sub_op.timestamp": datetime,
    }


def test_recursive_perspective_flattening():
    now = datetime.now(timezone.utc).replace(tzinfo=timezone.utc)

    # List with 1 element
    o = MyTestStruct(sub=MyTestSubStruct(y=[1], timestamp=now), timestamp=now)
    sub_id = MyTestSubStruct.id_generator.current()
    id = MyTestStruct.id_generator.current()

    assert isinstance(o.sub.psp_flatten(), list)
    sub_flat = o.sub.psp_flatten()
    for d in sub_flat:
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
    assert sub_flat == [
        {"id": str(sub_id), "y": 1, "x": 1, "timestamp": now},
    ]

    assert isinstance(o.psp_flatten(), list)
    o_flat = o.psp_flatten()
    for d in o_flat:
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
        d["sub.timestamp"] = datetime.fromisoformat(d["sub.timestamp"])
    assert o_flat == [
        {
            "id": str(id),
            "sub.id": str(sub_id),
            "sub.y": 1,
            "sub.x": 1,
            "sub.timestamp": now,
            "timestamp": now,
        },
    ]

    o = MyTestStruct(sub=MyTestSubStruct(y=[], z=float("nan"), timestamp=now), timestamp=now)
    assert isinstance(o.psp_flatten(), list)
    o_flat = o.psp_flatten()
    for d in o_flat:
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
        d["sub.timestamp"] = datetime.fromisoformat(d["sub.timestamp"])
    assert o_flat == [
        {
            "id": str(id + 1),
            "sub.id": str(sub_id + 1),
            "sub.x": 1,
            "sub.z": None,
            "sub.timestamp": now,
            "timestamp": now,
        },
    ]

    o = MyTestStruct(sub=MyTestSubStruct(y=[], z=float("inf"), timestamp=now), timestamp=now)
    assert isinstance(o.psp_flatten(), list)
    o_flat = o.psp_flatten()
    for d in o_flat:
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
        d["sub.timestamp"] = datetime.fromisoformat(d["sub.timestamp"])
    assert o_flat == [
        {
            "id": str(id + 2),
            "sub.id": str(sub_id + 2),
            "sub.x": 1,
            "sub.z": None,
            "sub.timestamp": now,
            "timestamp": now,
        },
    ]

    o = MyTestStruct(sub=MyTestSubStruct(y=[], z=float("-inf"), timestamp=now), timestamp=now)
    assert isinstance(o.psp_flatten(), list)
    o_flat = o.psp_flatten()
    for d in o_flat:
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
        d["sub.timestamp"] = datetime.fromisoformat(d["sub.timestamp"])
    assert o_flat == [
        {
            "id": str(id + 3),
            "sub.id": str(sub_id + 3),
            "sub.x": 1,
            "sub.z": None,
            "sub.timestamp": now,
            "timestamp": now,
        },
    ]

    # List with multiple elements
    o = MyTestStruct(sub=MyTestSubStruct(y=[1, 2, 3], timestamp=now), timestamp=now)

    assert isinstance(o.sub.psp_flatten(), list)
    sub_flat = o.sub.psp_flatten()
    for d in sub_flat:
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
    assert sub_flat == [
        {"id": str(sub_id + 4), "y": 1, "x": 1, "timestamp": now},
        {"id": str(sub_id + 4), "y": 2, "x": 1, "timestamp": now},
        {"id": str(sub_id + 4), "y": 3, "x": 1, "timestamp": now},
    ]

    assert isinstance(o.psp_flatten(), list)
    o_flat = o.psp_flatten()
    for d in o_flat:
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
        d["sub.timestamp"] = datetime.fromisoformat(d["sub.timestamp"])
    assert o_flat == [
        {
            "id": str(id + 4),
            "sub.id": str(sub_id + 4),
            "sub.y": 1,
            "sub.x": 1,
            "sub.timestamp": now,
            "timestamp": now,
        },
        {
            "id": str(id + 4),
            "sub.id": str(sub_id + 4),
            "sub.y": 2,
            "sub.x": 1,
            "sub.timestamp": now,
            "timestamp": now,
        },
        {
            "id": str(id + 4),
            "sub.id": str(sub_id + 4),
            "sub.y": 3,
            "sub.x": 1,
            "sub.timestamp": now,
            "timestamp": now,
        },
    ]

    # Empty list removed
    o = MyTestStruct(sub=MyTestSubStruct(y=[], timestamp=now), timestamp=now)
    assert isinstance(o.sub.psp_flatten(), list)
    sub_flat = o.sub.psp_flatten()
    for d in sub_flat:
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
    assert sub_flat == [
        {"id": str(sub_id + 5), "x": 1, "timestamp": now},
    ]

    assert isinstance(o.psp_flatten(), list)
    o_flat = o.psp_flatten()
    for d in o_flat:
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
        d["sub.timestamp"] = datetime.fromisoformat(d["sub.timestamp"])
    assert o_flat == [
        {
            "id": str(id + 5),
            "sub.id": str(sub_id + 5),
            "sub.x": 1,
            "sub.timestamp": now,
            "timestamp": now,
        },
    ]

    # Non json serializable object passed
    o = MyTestStruct(ncsp_sub=object(), timestamp=now)
    assert isinstance(o.psp_flatten(), list)
    o_flat = o.psp_flatten()
    for d in o_flat:
        d["timestamp"] = datetime.fromisoformat(d["timestamp"])
    assert o_flat == [
        {
            "id": str(id + 6),
            "ncsp_sub": "",
            "timestamp": now,
        },
    ]


def test_exclude_columns_schema():
    res = MyTestStruct.psp_schema()
    assert set(res.keys()) == {"z", "sub.y", "sub.x", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}

    res = MyTestStruct.psp_schema({"z"})
    assert set(res.keys()) == {"sub.y", "sub.x", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}

    res = MyTestStruct.psp_schema({"sub"})
    assert set(res.keys()) == {"z", "id", "timestamp"}

    res = MyTestStruct.psp_schema({"sub": {"x"}})
    assert set(res.keys()) == {"z", "sub.y", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}

    res = MyTestStruct.psp_schema({"sub": {"x"}, "z": True})
    assert set(res.keys()) == {"sub.y", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}


def test_pyarrow_conversion():
    now = datetime.now(timezone.utc)
    now_date = now.date()
    schema = MyPyArrowStruct.psp_schema()
    arrow_schema = psp_schema_to_arrow_schema(schema)
    date_key_set = set()
    date_key_set.add("d")
    datetime_key_set = set()
    datetime_key_set.add("timestamp")

    o = MyPyArrowStruct(timestamp=now, d=now_date)
    table = create_pyarrow_table(None, [o], arrow_schema, date_key_set)
    df = pl.from_arrow(table)
    assert df.head()["d"][0] == now_date

    #  Check nones are handled correctly
    o = MyPyArrowStruct(timestamp=None, d=None)
    table = create_pyarrow_table(None, [o], arrow_schema, date_key_set)
    df = pl.from_arrow(table)
    assert df.head()["d"][0] is None
    assert df.head()["timestamp"][0] is None


@pytest.mark.parametrize("use_external_perspective", [True, False])
def test_MountPerspectiveTables(use_external_perspective):
    module = MountPerspectiveTables(
        tables={"exclude": ["exclude_channel"]},
        limits={"limit_channel": 5},
        indexes={"index_channel": "id", "dict_channel": "id", "exclude_channel": None},
        layouts={
            "Test Layout": '{"sizes":[1],"detail":{"main":{"type":"split-area","orientation":"horizontal","children":[{"type":"split-area","orientation":"vertical","children":[{"type":"tab-area","widgets":["PERSPECTIVE_GENERATED_ID_1"],"currentIndex":0},{"type":"tab-area","widgets":["PERSPECTIVE_GENERATED_ID_4"],"currentIndex":0}],"sizes":[0.5,0.5]},{"type":"split-area","orientation":"vertical","children":[{"type":"tab-area","widgets":["PERSPECTIVE_GENERATED_ID_3"],"currentIndex":0},{"type":"tab-area","widgets":["PERSPECTIVE_GENERATED_ID_5"],"currentIndex":0}],"sizes":[0.5,0.5]}],"sizes":[0.5,0.5]}},"mode":"globalFilters","viewers":{"PERSPECTIVE_GENERATED_ID_1":{"plugin":"Datagrid","plugin_config":{"columns":{},"editable":false,"scroll_lock":true},"settings":false,"theme":"Pro Dark","group_by":["id"],"split_by":[],"columns":["timestamp","x","y"],"filter":[],"sort":[["timestamp","desc"]],"expressions":[],"aggregates":{"timestamp":"last","x":"last","id":"last","y":"last"},"master":false,"name":"basket","table":"basket","linked":false},"PERSPECTIVE_GENERATED_ID_4":{"plugin":"Datagrid","plugin_config":{"columns":{},"editable":false,"scroll_lock":true},"settings":false,"theme":"Pro Dark","group_by":["id"],"split_by":[],"columns":["timestamp","x","y"],"filter":[],"sort":[["timestamp","desc"]],"expressions":[],"aggregates":{"timestamp":"last","x":"last","id":"last","y":"last"},"master":false,"name":"example_list","table":"example_list","linked":false},"PERSPECTIVE_GENERATED_ID_3":{"plugin":"Datagrid","plugin_config":{"columns":{},"editable":false,"scroll_lock":true},"settings":false,"theme":"Pro Dark","group_by":["id"],"split_by":[],"columns":["timestamp","x","y"],"filter":[],"sort":[["timestamp","desc"]],"expressions":[],"aggregates":{"id":"last","x":"last","timestamp":"last","y":"last"},"master":false,"name":"example","table":"example","linked":false},"PERSPECTIVE_GENERATED_ID_5":{"plugin":"Datagrid","plugin_config":{"columns":{},"editable":false,"scroll_lock":true},"settings":false,"theme":"Pro Dark","group_by":["id"],"split_by":[],"columns":["timestamp","x","y"],"filter":[],"sort":[["timestamp","desc"]],"expressions":[],"aggregates":{"id":"last","timestamp":"last","x":"last","y":"last"},"master":false,"name":"never_ticks","table":"never_ticks","linked":false}}}'  # noqa: E501
        },
        update_interval=timedelta(seconds=0.5),
    )
    if use_external_perspective:
        module.perspective_field = "my_perspective"

    h = GatewayTestHarness(
        test_channels=[GWC.test_channel, GWC.list_channel, GWC.dict_channel, GWC.dict_enum_channel, GWC.index_channel, GWC.limit_channel],
        test_dynamic_keys={"dict_channel": ["test_key"]},
    )

    now = datetime.now(timezone.utc).replace(tzinfo=timezone.utc)

    # List with 1 element
    o = MyTestStruct(sub=MyTestSubStruct(y=[1], timestamp=now), timestamp=now)

    for _ in range(10):
        h.send(GWC.test_channel, o)
        h.send(GWC.list_channel, [o])
        h.send(GWC.dict_channel, {"test_key": o})
        h.send(GWC.dict_enum_channel, {MyTestEnum.a: o})
        h.send(GWC.limit_channel, o)
        h.send(GWC.index_channel, o)
    h.assert_ticked(GWC.test_channel, 10)
    h.assert_ticked(GWC.list_channel, 10)
    h.assert_ticked((GWC.dict_channel, "test_key"), 10)
    h.assert_ticked((GWC.dict_enum_channel, MyTestEnum.a), 10)
    h.assert_ticked(GWC.limit_channel, 10)
    h.assert_ticked(GWC.index_channel, 10)

    h.delay(2 * module.update_interval)  # So that the buffer is flushed
    channels = GWC()
    gateway = Gateway(modules=[h, module], channels=channels)
    csp.run(gateway.graph, starttime=datetime(2023, 1, 1), endtime=timedelta(5))

    # The graph runs, which is a good sign. Now look at the contents of the tables more closely

    if use_external_perspective:
        psp = channels.my_perspective  # Confirm we use the manager from the channels in the tests that follow
    else:
        psp = module._server

    assert sorted(psp.new_local_client().get_hosted_table_names()) == sorted(
        [
            GWC.test_channel,
            GWC.list_channel,
            GWC.dict_channel,
            GWC.dict_enum_channel,
            GWC.index_channel,
            GWC.limit_channel,
        ]
    )

    def table_len(name):
        return len(psp.new_local_client().open_table(name).view().to_json())

    assert table_len(GWC.test_channel) == 10
    assert table_len(GWC.list_channel) == 10

    assert table_len(GWC.dict_channel) == 1  # Because of index
    assert table_len(GWC.dict_enum_channel) == 10
    assert table_len(GWC.limit_channel) == 5
    assert table_len(GWC.index_channel) == 1


@pytest.mark.parametrize("exclude_columns", [True, False])
def test_MountPerspectiveTables_exclude_columns(exclude_columns):
    if exclude_columns:
        excluded_table_columns = {"test_channel": {"z"}, "list_channel": {"sub": {"y"}}, "dict_channel": {"sub": {"y": True, "x": True}}}
    else:
        excluded_table_columns = {}

    module = MountPerspectiveTables(
        excluded_table_columns=excluded_table_columns,
        update_interval=timedelta(seconds=0.5),
    )

    h = GatewayTestHarness(
        test_channels=[GWC.test_channel, GWC.list_channel, GWC.dict_channel, GWC.exclude_channel],
        test_dynamic_keys={"dict_channel": ["test_key"]},
    )

    now = datetime.now(timezone.utc).replace(tzinfo=timezone.utc)

    # List with 1 element
    o = MyTestStruct(sub=MyTestSubStruct(y=[1], timestamp=now), timestamp=now)

    h.send(GWC.test_channel, o)
    h.send(GWC.list_channel, [o])
    h.send(GWC.dict_channel, {"test_key": o})
    h.send(GWC.exclude_channel, o)

    h.delay(2 * module.update_interval)  # So that the buffer is flushed
    channels = GWC()
    gateway = Gateway(modules=[h, module], channels=channels)
    csp.run(gateway.graph, starttime=datetime(2023, 1, 1), endtime=timedelta(5))

    def table_columns(name):
        return set(module._server.new_local_client().open_table(name).columns())

    def table_len(name):
        return len(module._server.new_local_client().open_table(name).view().to_json())

    if exclude_columns:
        assert table_columns(GWC.test_channel) == {"sub.y", "sub.x", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}
        assert table_columns(GWC.list_channel) == {"z", "sub.x", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}
        assert table_columns(GWC.dict_channel) == {"z", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}

    else:
        assert table_columns(GWC.test_channel) == {"z", "sub.y", "sub.x", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}
        assert table_columns(GWC.list_channel) == {"z", "sub.y", "sub.x", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}
        assert table_columns(GWC.dict_channel) == {"z", "sub.y", "sub.x", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}

    assert table_columns(GWC.exclude_channel) == {"z", "sub.y", "sub.x", "sub.z", "id", "timestamp", "sub.id", "sub.timestamp"}

    assert table_len(GWC.test_channel) == 1
    assert table_len(GWC.list_channel) == 1
    assert table_len(GWC.dict_channel) == 1
    assert table_len(GWC.exclude_channel) == 1


def test_exclude_unused_tables():
    channels = GWCUnused()
    psp_module = MountPerspectiveTables()

    class MyTestModule(GatewayModule):
        def connect(self, channels):
            channels.set_channel(GWCUnused.a, csp.const(MyTestStruct()))
            channels.set_channel(GWCUnused.c, csp.const([MyTestStruct()]))
            channels.set_channel(GWCUnused.e, csp.const(MyTestStruct()), MyTestEnum.a)

    gateway = Gateway(modules=[psp_module, MyTestModule()], channels=channels)

    # Build gateway - this will wire unused channels with null_ts
    csp.run(gateway.graph, starttime=datetime(2023, 1, 1), endtime=timedelta(5))

    # Mock out startup
    gateway.output = MagicMock()
    gateway.graph_build_failed = False
    gateway.running = True
    gateway.output.is_alive.return_value = True

    # Build web to run perspective table pruning
    gateway._build_web(ui=True, timeout=1, _in_test=True)

    # Assert that tables were pruned
    assert psp_module._unused_tables == ["b", "d", "f"]



import typing
from datetime import datetime, timezone

from atomic_counter import Counter

if typing.TYPE_CHECKING:
    from csp_gateway.server import GatewayModule


def get_counter(kind: "GatewayModule"):
    if not hasattr(get_counter, "id_map"):
        get_counter.map = {}
    if kind not in get_counter.map:
        nowish = datetime.now(timezone.utc)
        base = datetime(nowish.year, nowish.month, nowish.day)
        get_counter.map[kind] = Counter(int(base.timestamp()) * 1_000_000_000)
    return get_counter.map[kind]



import csv
import logging
from dataclasses import dataclass
from datetime import datetime
from enum import Enum, auto
from typing import Any, Callable, Dict, List, Optional, TypeVar

import csp
import orjson
import pyarrow.parquet as pq
from csp.impl.pushadapter import PushInputAdapter
from csp.impl.types.container_type_normalizer import ContainerTypeNormalizer
from csp.impl.wiring import py_push_adapter_def
from pydantic import TypeAdapter
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

__all__ = (
    "FileDropType",
    "FileDropAdapterConfiguration",
    "filedrop_adapter_def",
)


log = logging.getLogger(__name__)


class FileDropType(Enum):
    CUSTOM = auto()
    CSV = auto()
    JSON = auto()
    PARQUET = auto()


@dataclass
class FileDropAdapterConfiguration:
    """Configuration for the filedrop push adapter"""

    # Path to the directory to monitor
    dir_path: str
    # Format of files to expect to load properly i.e parquet, json, csv
    filedrop_type: FileDropType
    # List of extensions to filter, empty list means all extensions are allowed
    extensions: List[str]
    # custom loader for loading files into list of struct like data
    loader: Optional[Callable[[str], List[Any]]]
    # deserialize each data point to data type expected by channel type
    deserializer: Optional[Callable[[Any], Any]]
    # Extra args to the type adapter deserializer
    type_adapter_args: Dict[str, Any]


class FileReaderBase:
    """The base file reader that reads data from files and generates structs"""

    def __init__(self, config: FileDropAdapterConfiguration, ts_typ: object):
        self.extensions = config.extensions
        if hasattr(config, "type_adapter_args"):
            self.context = config.type_adapter_args
        else:
            self.context = {}
        if hasattr(config, "deserializer") and config.deserializer:
            self.deserializer = config.deserializer
        else:
            normalized_type = ContainerTypeNormalizer.normalize_type(ts_typ)
            type_adapter = TypeAdapter(normalized_type)

            def deserialize_tick(data, type_adapter=type_adapter, context=self.context):
                return type_adapter.validate_python(data, context=context)

            self.deserializer = deserialize_tick

    def read(self, src_path: str) -> object:
        """Generator to return stucts from a filepath"""

        should_read = True
        if self.extensions:
            if not any([src_path.endswith(suffix) for suffix in self.extensions]):
                should_read = False
        if should_read:
            dicts = self.read_impl(src_path)
            structs = [self.deserializer(d) for d in dicts]
            for s in structs:
                yield s

    def read_impl(self, src_path: str) -> List[dict]:
        """File type specific implementation"""

        raise Exception(f"read not implemented for {self}")


class FileReaderCsv(FileReaderBase):
    """File reader for json file type"""

    def read_impl(self, src_path: str) -> List[Any]:
        data = []
        with open(src_path, "r") as f:
            reader = csv.DictReader(f)
            for row in reader:
                data.append(row)
        return data


class FileReaderJson(FileReaderBase):
    """File reader for json file type"""

    def read_impl(self, src_path: str) -> List[Any]:
        with open(src_path, "rb") as f:
            data = orjson.loads(f.read())
        if isinstance(data, list):
            res = data
        else:
            res = [data]
        return res


class FileReaderParquet(FileReaderBase):
    """File reader for parquet file type"""

    def read_impl(self, src_path: str) -> List[Any]:
        table = pq.read_table(src_path)
        return table.to_pylist()


class FileReaderCustom(FileReaderBase):
    """File reader for a custom file loader"""

    def __init__(self, config: FileDropAdapterConfiguration, ts_typ: object):
        super().__init__(config, ts_typ)
        self._loader = config.loader

    def read_impl(self, src_path: str) -> List[Any]:
        return self._loader(src_path)


class EventHandlerCustom(FileSystemEventHandler):
    def __init__(self, adapter: PushInputAdapter, file_reader: FileReaderBase):
        self.file_reader = file_reader
        self.adapter = adapter
        self._created = set()
        self._opened = set()
        self._modified = set()
        self._closed = set()

    def on_created(self, event: FileSystemEvent):
        self._created.add(event.src_path)

    def on_opened(self, event: FileSystemEvent):
        if event.src_path in self._created:
            self._created.remove(event.src_path)
            self._opened.add(event.src_path)

    def on_modified(self, event: FileSystemEvent):
        if event.src_path in self._opened:
            self._opened.remove(event.src_path)
            self._modified.add(event.src_path)

    def on_closed(self, event: FileSystemEvent):
        if event.src_path in self._modified:
            self._modified.remove(event.src_path)
            file_path = event.src_path
            try:
                for data in self.file_reader.read(file_path):
                    self.adapter.push_tick(data)
            except Exception as e:
                log.error(f"Failed to read data from {file_path} with exception: {e}, skipping")


T = TypeVar("T")


class _FileDropImpl(PushInputAdapter):
    FILEREADER_MAP = {
        FileDropType.CSV: FileReaderCsv,
        FileDropType.JSON: FileReaderJson,
        FileDropType.PARQUET: FileReaderParquet,
        FileDropType.CUSTOM: FileReaderCustom,
    }

    def __init__(self, config: FileDropAdapterConfiguration, ts_typ: T):
        # NOTE ts_typ is assumed to be a List["Y"] type where "Y" is the actual type
        self.dir_path = config.dir_path
        self.observer = Observer()
        reader = self.FILEREADER_MAP[config.filedrop_type]
        file_reader = reader(config, ts_typ[0])
        self.event_handler = EventHandlerCustom(self, file_reader)
        self.observer.schedule(self.event_handler, self.dir_path, recursive=False)

    def start(self, starttime: datetime, endtime: datetime):
        self.observer.start()
        self.observer_started = True

    def stop(self):
        if self.observer_started:
            self.observer.stop()
            self.observer.join()


# NOTE: Ref: https://github.com/Point72/csp/issues/569
# Instead of passing the actual type T we want as ts_typ, we have to pass in List[T]
# This is due to a type normalization bug in csp, where it converts Dict[K, T] to dict, i.e removing the type annotation hints
# The type annotations are needed to properly deserialize the raw data using the pydantic type adapters.
# Further we can only take the object type as output due to issues caused by using List[T] instead of T and bypassing
# the type normalization process in csp. The user of the adapter, needs to keep track of the type they pass in ts_type,
# and cast the ts[object] to the required ts[T], they want using csp.appy
_filedrop_adapter_def = py_push_adapter_def(
    "_filedrop_adapter_def",
    _FileDropImpl,
    csp.ts[object],
    config=FileDropAdapterConfiguration,
    ts_typ=List[T],
)


@csp.graph
def filedrop_adapter_def(config: FileDropAdapterConfiguration, ts_typ: "T") -> csp.ts["T"]:
    object_data = _filedrop_adapter_def(config=config, ts_typ=[ts_typ])
    data = csp.apply(object_data, lambda x: x, ts_typ)
    return data



import time
from datetime import datetime, timedelta, timezone
from unittest import mock

import csp
import orjson
import pytest

from csp_gateway import ReadWriteMode, State
from csp_gateway.server import (
    ChannelSelection,
    EncodedEngineCycle,
    Mirror,
    ReplayEngineJSON,
)
from csp_gateway.server.gateway.csp.channels import _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD
from csp_gateway.testing.shared_helpful_classes import (
    MyGateway,
    MyGatewayChannels,
    MySetModule,
    MyStruct,
)

from .kafka.test_kafka import create_kafka_engine_replay


@mock.patch("csp_gateway.ReplayEngineKafka", autospec=True)
def test_mirror_same_channels(mock_object):
    requires = ChannelSelection(include=["test_channel1"])
    selection = ChannelSelection(include=["test_channel2"])

    mirror = Mirror(mirror_source=mock_object, requires=requires, selection=selection)

    assert mirror.mirror_source.requires == requires
    assert mirror.mirror_source.selection == selection


@mock.patch("csp.adapters.kafka.KafkaAdapterManager", autospec=True)
@pytest.mark.parametrize("read_write_mode", [ReadWriteMode.READ, ReadWriteMode.READ_AND_WRITE])
def test_kafka_engine_replay_read_and_write_read(mock_object, read_write_mode):
    mock_publish = mock.MagicMock()

    def mock_subscribe(*a, **kw):
        return csp.null_ts(EncodedEngineCycle)

    mock_object.return_value.subscribe.side_effect = mock_subscribe
    mock_object.return_value.publish = mock_publish

    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    kafka_engine_replay_write = create_kafka_engine_replay(ChannelSelection.model_validate(channels), read_write_mode)

    mirror = Mirror(
        mirror_source=kafka_engine_replay_write,
        selection=kafka_engine_replay_write.selection,
    )
    gateway_writing = MyGateway(modules=[setter, mirror, setter], channels=MyGatewayChannels())
    csp.run(gateway_writing.graph, starttime=datetime.now(timezone.utc), endtime=timedelta(seconds=5))
    assert mock_publish.call_count == 0


@pytest.mark.parametrize("read_write_mode", [ReadWriteMode.WRITE, ReadWriteMode.READ])
def test_overwrite_json(read_write_mode, tmpdir):
    file = str(tmpdir.join("json_test_data.json"))
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    setup_json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.WRITE,
    )
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    gateway = MyGateway(modules=[setup_json_module, setter], channels=MyGatewayChannels())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    with open(file, "r") as json_file:
        assert len(list(json_file)) == 1

    # start_writing doesnt matter since Mirror is only in READ mode
    json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        start_writing=datetime(2020, 12, 12),
        read_write_mode=read_write_mode,
        overwrite_if_writing=True,
    )
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0), timedelta(days=365)),
        my_data2=csp.const(MyStruct(foo=2.0), timedelta(days=365)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)], timedelta(days=365)),
    )
    mirror = Mirror(mirror_source=json_module, selection=json_module.selection)
    new_gateway = MyGateway(modules=[mirror, setter], channels=MyGatewayChannels())
    csp.run(new_gateway.graph, starttime=datetime(2020, 1, 1))
    with open(file, "r") as json_file:
        items = list(json_file)
        assert len(items) == 1
        json_dict = orjson.loads(items[0])
        actual_date = datetime.fromisoformat(json_dict[_CSP_ENGINE_CYCLE_TIMESTAMP_FIELD]).replace(tzinfo=None)
        assert actual_date == datetime(2020, 1, 1)


@pytest.mark.parametrize("by_key", [True, False])
def test_mirror_state(by_key, tmpdir):
    file = str(tmpdir.join("json_test_data.json"))
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    setup_json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        read_write_mode=ReadWriteMode.WRITE,
    )
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    gateway = MyGateway(modules=[setup_json_module, setter], channels=MyGatewayChannels())
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    with open(file, "r") as json_file:
        assert len(list(json_file)) == 1

    json_module = ReplayEngineJSON(
        selection=channels,
        filename=file,
        start_writing=datetime(2020, 12, 12),
        read_write_mode=ReadWriteMode.READ,
        overwrite_if_writing=True,
    )
    state_channels = {
        MyGatewayChannels.my_channel: "id",
        MyGatewayChannels.my_list_channel: "id",
    }
    mirror = Mirror(
        mirror_source=json_module,
        selection=json_module.selection,
        state_channels=state_channels,
    )
    new_gateway = MyGateway(modules=[mirror], channels=MyGatewayChannels())
    new_gateway.start(block=False, realtime=True, rest=False, starttime=datetime(2020, 1, 1))
    try:
        time.sleep(0.5)
        state = new_gateway.channels.state("my_channel")
        assert isinstance(state, State)
        assert len(state.query()) == 1
        state = new_gateway.channels.state("my_list_channel")
        assert isinstance(state, State)
        assert len(state.query()) == 2
    finally:
        with mock.patch("os._exit"), mock.patch("os.kill"):
            new_gateway.stop()



import time

import csp
from csp import ts

from csp_gateway import GatewayModule, get_thread

__all__ = (
    "NeverDieModule",
    "CspDieModule",
    "LongStartModule",
)


def _never_die():
    while True:
        try:
            print("here")
            time.sleep(1)
        except:  # noqa: E722
            ...


class NeverDieModule(GatewayModule):
    def connect(self, *args, **kwargs) -> None:
        self._thread = get_thread(target=_never_die)
        self._thread.start()


class CspDieModule(GatewayModule):
    count: int = 5
    channel_to_access: str = "example"
    attribute_to_access: str = "blerg"

    def connect(self, channels) -> None:
        self._tick(channels.get_channel(self.channel_to_access))

    @csp.node
    def _tick(self, thing: ts[object]):
        with csp.state():
            s_counter = 0
        if csp.ticked(thing):
            s_counter += 1
            if s_counter >= self.count:
                print("running getter to trigger problem")
                getattr(thing, self.attribute_to_access)


class LongStartModule(GatewayModule):
    sleep: int = 5
    channel_to_access: str = "example"

    def connect(self, channels) -> None:
        time.sleep(self.sleep)
        self._tick(channels.get_channel(self.channel_to_access))

    @csp.node
    def _tick(self, thing: ts[object]):
        with csp.start():
            time.sleep(self.sleep)
        if csp.ticked(thing):
            print("do nothing")



import logging
import os
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union, get_args, get_origin

import csp
from ccflow import BaseModel
from csp.impl.types.container_type_normalizer import ContainerTypeNormalizer
from csp.impl.types.tstype import isTsType
from pydantic import Field, model_validator

from csp_gateway.server import (
    GatewayChannels,
    GatewayModule,
)

from .adapter import FileDropAdapterConfiguration, FileDropType, filedrop_adapter_def

__all__ = (
    "ReadFileDropConfiguration",
    "ReadFileDrop",
)

log = logging.getLogger(__name__)


class ReadFileDropConfiguration(BaseModel):
    """The configuration of a filedrop adapter for a directory and filetype"""

    dir_path: Path = Field(description="Directory to monitor for new files")
    channel_name: str = Field(description="Name of the channel to send the structs to")
    filedrop_type: Union[FileDropType, str] = Field(description="The type of files to expect and accordingly read i.e. parquet, json, etc")
    extensions: List[str] = Field(default=[], description="List of extensions to decide which files to read, empty list means all extensions")
    loader: Optional[Callable[[str], List[Any]]] = Field(default=None, description="custom loader for loading files into list of struct like data")
    deserializer: Optional[Callable[[Any], Any]] = Field(
        default=None, description="deserialize each data point to data type expected by channel type"
    )
    subscribe_with_struct_id: bool = Field(
        default=False,
        description=("If False, replaces the id field on GatewayStructs from files with one autogenerated by the current Gateway."),
    )
    subscribe_with_struct_timestamp: bool = Field(
        default=False,
        description=("If False, replaces the timestamp field on the GatewayStruct with a timestamp autogenerated by the current Gateway."),
    )

    @model_validator(mode="after")
    def check_filedrop_type(self) -> "ReadFileDropConfiguration":
        if isinstance(self.filedrop_type, str):
            try:
                self.filedrop_type = FileDropType[self.filedrop_type]
            except KeyError:
                raise ValueError(f"{self.filedrop_type} is not a valid FileDropType")
        if self.filedrop_type == FileDropType.CUSTOM and self.loader is None:
            raise ValueError("loader must be set if filedrop_type is CUSTOM")
        return self


T = TypeVar("T")
K = TypeVar("K")


class ReadFileDrop(GatewayModule):
    """The module to read data from files dropped in specific directories and send them as structs to specific channels"""

    configs: List[ReadFileDropConfiguration] = Field(description="List of configs for the directories to monitor")

    @csp.node
    def handle_list_basket(self, data: csp.ts[List[T]], list_size: int) -> csp.OutputBasket(List[csp.ts[T]], shape="list_size"):
        if csp.ticked(data):
            return data

    @csp.node
    def handle_dict_basket(self, data: csp.ts[Dict[K, T]], dict_keys: List[K]) -> csp.OutputBasket(Dict[K, csp.ts[T]], shape="dict_keys"):
        if csp.ticked(data):
            return data

    def connect(self, channels: GatewayChannels):
        channel_data = {}
        channel_basket_types = {}
        for config in self.configs:
            # ensure that directories exist
            os.makedirs(config.dir_path, exist_ok=True)
            channel_data[config.channel_name] = []
        for config in self.configs:
            context = {
                "force_new_id": not config.subscribe_with_struct_id,
                "force_new_timestamp": not config.subscribe_with_struct_timestamp,
            }
            adapter_config = FileDropAdapterConfiguration(
                dir_path=str(config.dir_path),
                filedrop_type=config.filedrop_type,
                extensions=config.extensions,
                type_adapter_args=context,
                loader=config.loader,
                deserializer=config.deserializer,
            )
            channel_type = channels.get_outer_type(config.channel_name)
            if isTsType(channel_type):
                non_ts_type = channel_type.typ
                channel_basket_types[config.channel_name] = ""
            else:
                normalized_type = ContainerTypeNormalizer.normalize_type(channel_type)
                if get_origin(normalized_type) is list:
                    inner_type = get_args(normalized_type)[0]
                    if not isTsType(inner_type):
                        raise ValueError(f"Channel type for {config.channel_name} should be of the form List[TsType], got: {channel_type}")
                    non_ts_type = List[inner_type.typ]
                    channel_basket_types[config.channel_name] = "list"
                elif get_origin(normalized_type) is dict:
                    key_type, inner_type = get_args(normalized_type)
                    if not isTsType(inner_type):
                        raise ValueError(f"Channel type for {config.channel_name} should be of the form Dict[KeyType, TsType], got: {channel_type}")
                    non_ts_type = Dict[key_type, inner_type.typ]
                    channel_basket_types[config.channel_name] = "dict"
                else:
                    raise Exception(f"Channel type cannot be handled: {channel_type}")
            data = filedrop_adapter_def(config=adapter_config, ts_typ=non_ts_type)
            channel_data[config.channel_name].append(data)
        for channel_name, data_list in channel_data.items():
            data = csp.flatten(data_list)
            if channel_basket_types[channel_name] == "list":
                channels.set_channel(channel_name, self.handle_list_basket(data, list_size=channels.dynamic_keys()[channel_name]))
            elif channel_basket_types[channel_name] == "dict":
                channels.set_channel(channel_name, self.handle_dict_basket(data, dict_keys=channels.dynamic_keys()[channel_name]))
            else:
                channels.set_channel(channel_name, data)



import json
import logging
from collections import deque
from datetime import timedelta
from typing import Dict, List, Literal, Optional

import csp
import opsgenie_sdk
from csp import ts
from pydantic import Field, PrivateAttr, model_validator

from csp_gateway.server import ChannelSelection, GatewayModule
from csp_gateway.server.modules.logging.util import (
    MonitoringEvent,
    MonitoringLevelMapping,
    MonitoringMetric,
    OpsGenieLevel,
)

log = logging.getLogger(__name__)

# define TRACE level logging below DEBUG to investigate OpsGenie submission
# without polluting DEBUG level file log
TRACE_LEVEL = 5
logging.addLevelName(TRACE_LEVEL, "TRACE")


def trace(self, message, *args, **kwargs):
    if self.isEnabledFor(TRACE_LEVEL):
        self._log(TRACE_LEVEL, message, args, **kwargs)


logging.Logger.trace = trace

__all__ = ("PublishOpsGenie",)


class PublishOpsGenie(GatewayModule):
    """
    PublishOpsGenie class is responsible for interfacing with OpsGenie to send alerts and
    manage heartbeats and heartbeat lags. It includes configurations for API keys, tags,
    delays, and heartbeat settings.
    """

    ops_api_key: str = Field(description="The API key for OpsGenie.")

    # None of the channels are required
    requires: Optional[ChannelSelection] = Field(default=[], description="List of required channels.")
    events_channel: Optional[str] = Field(default=None, description="Channel for events.")
    metrics_channel: Optional[str] = Field(default=None, description="Channel for metrics.")
    _api_client: opsgenie_sdk.api_client.ApiClient = PrivateAttr()

    ops_tags: Optional[Dict[str, str]] = Field(default=None, description="Tags to be included with OpsGenie alerts.")
    ops_async_delay_sec: Optional[float] = Field(default=5.0, description="Delay in seconds for asynchronous operations.")
    ops_sync_delay_sec: Optional[float] = Field(default=5.0, description="Delay in seconds for synchronous operations.")

    # heartbeat related
    ops_heartbeat_name: str = Field(description="The name of the heartbeat.")
    ops_heartbeat_interval: Optional[int] = Field(default=1, description="The interval at which heartbeats are sent.")
    ops_heartbeat_interval_unit: Optional[Literal["minutes", "hours", "days"]] = Field(
        default="minutes", description="The unit of the heartbeat interval."
    )
    ops_heartbeat_alert_level: Optional[OpsGenieLevel] = Field(default=OpsGenieLevel.P2, description="The alert level for heartbeat issues.")
    ops_heartbeat_metric_name: Optional[str] = Field(default="heartbeat", description="The metric name for heartbeat.")
    ops_alert_min_level: Optional[OpsGenieLevel] = Field(
        default=OpsGenieLevel.P3,
        description="The minimum alert level for OpsGenie alerts.",
    )
    ops_alias_tags: Dict[str, List[str]] = Field(default={}, description="Tags to be used as OpsGenie alis for event_agregation")
    ops_alias_separator: str = Field(default=":", description="Separator to build OpsGenie alias.")
    ops_category_tag: str = Field(default="event_group", description="Tag that identifies event type.")

    @model_validator(mode="before")
    def check_metrics_channel_and_heartbeat(cls, values):
        if values.get("metrics_channel") and not values.get("ops_heartbeat_name"):
            raise ValueError("'ops_heartbeat_name' is required if 'metrics_channel' is provided")
        return values

    def _init_heartbeat(self, heartbeat_api):
        # we specifically want to run heartbeat initialization in async_req=False with timeout
        # then each ping will be run in async_req=True mode
        payload = opsgenie_sdk.CreateHeartbeatPayload(
            name=self.ops_heartbeat_name,
            description=f"{self.ops_heartbeat_name} generated by csp_gateway",
            interval=self.ops_heartbeat_interval,
            interval_unit=self.ops_heartbeat_interval_unit,
            enabled=True,
            alert_message=(
                f"{self.ops_heartbeat_name} heartbeat not received in " + f"{self.ops_heartbeat_interval} {self.ops_heartbeat_interval_unit}"
            ),
            alert_tags=[f"{k}:{v}" for k, v in self.ops_tags.items()],
            alert_priority=self.ops_heartbeat_alert_level.value,
        )
        try:
            # heartbeats must have unique name
            # check if the heartbeat with this name already exists
            res = heartbeat_api.get_heartbeat(
                self.ops_heartbeat_name,
                async_req=False,
                _request_timeout=self.ops_sync_delay_sec,
            )
            log.info(
                "Heartbeat '%s' already exists: %s",
                self.ops_heartbeat_name,
                res.data.to_dict(),
            )
            # update heartbeat in case if payload parameters changed (i.e. tags, intervals, etc.)
            heartbeat_api.update_heartbeat(
                name=self.ops_heartbeat_name,
                update_heartbeat_payload=payload,
                async_req=False,
                _request_timeout=self.ops_sync_delay_sec,
            )
            log.info(
                "Updated heartbeat '%s' with the payload: %s",
                self.ops_heartbeat_name,
                res.data.to_dict(),
            )
        except opsgenie_sdk.exceptions.ApiException as e:
            # try to create a heartbeat if one does not exist
            message = json.loads(e.body)
            log.info(
                "Failed to retrieve heartbeat '%s': %s",
                self.ops_heartbeat_name,
                message,
            )
            res = heartbeat_api.create_heartbeat(
                create_heartbeat_payload=payload,
                async_req=False,
                _request_timeout=self.ops_sync_delay_sec,
            )
            log.info(
                "Created heartbeat '%s' with the payload: %s",
                self.ops_heartbeat_name,
                res.data.to_dict(),
            )

    def connect(self, channels):
        """
        Channels to be connected to graph
        """
        # this just creates an object and does not perform any calls or connections to OpsGenie
        config = opsgenie_sdk.configuration.Configuration()
        config.api_key["Authorization"] = self.ops_api_key
        self._api_client = opsgenie_sdk.api_client.ApiClient(configuration=config)

        if self.events_channel:
            events = channels.get_channel(self.events_channel)
            self._publish_alerts(events)

        if self.metrics_channel:
            metrics = channels.get_channel(self.metrics_channel)
            self._publish_heartbeat(metrics)

    @csp.node
    def _publish_heartbeat(self, data: ts[List[MonitoringMetric]]):
        with csp.alarms():
            alarm = csp.alarm(bool)

        with csp.state():
            s_alarm_scheduled = False
            # s_heartbeat queue to store and process async replies from OpsGenie
            s_heartbeat_queue = deque()
            s_heartbeat_api = opsgenie_sdk.HeartbeatApi(api_client=self._api_client)

        with csp.start():
            # init heartbeat object in synchronous mode
            self._init_heartbeat(heartbeat_api=s_heartbeat_api)

        if csp.ticked(data):
            for metric in data:
                log.trace("Received metric %s", metric.metric)
                if metric.metric == self.ops_heartbeat_metric_name:
                    try:
                        # heartbeat received, send ping
                        log.trace("Processing heartbeat %s", metric.to_dict())
                        res = s_heartbeat_api.ping(self.ops_heartbeat_name, async_req=True)
                        # append to the asynch queue to check status later
                        s_heartbeat_queue.append((csp.now(), res))
                        if not s_alarm_scheduled:
                            log.trace(
                                "Setting alarm to check heartbeat queue in %s seconds",
                                self.ops_async_delay_sec,
                            )
                            csp.schedule_alarm(alarm, timedelta(seconds=self.ops_async_delay_sec), True)
                            s_alarm_scheduled = True
                    except opsgenie_sdk.exceptions.ApiException as e:
                        log.error(
                            "Failed to send heartbeat for metric %s : %s",
                            metric.to_dict(self.ops_tags),
                            e,
                        )

        if csp.ticked(alarm):
            # keep processing the first element in the queue until the queue is either empty
            # or the timestamp of the first element is less than ops_async_delay_sec in the past
            while s_heartbeat_queue and (csp.now() - s_heartbeat_queue[0][0]).total_seconds() >= self.ops_async_delay_sec:
                log.trace(
                    "Processing heartbeat queue, %s object(s) remaining",
                    len(s_heartbeat_queue),
                )
                (_, res) = s_heartbeat_queue.popleft()
                if res.ready():
                    log.trace("Heartbeat submission successful")
                else:
                    log.error(
                        "Heartbeat submission is not ready in %s seconds",
                        self.ops_async_delay_sec,
                    )
            log.trace(
                "Done processing heartbeat queue, %s object(s) remaining",
                len(s_heartbeat_queue),
            )
        # if more items remaining reschedule the alarm
        if s_heartbeat_queue:
            log.trace(
                "Resetting alarm to check heartbeat queue in %s seconds",
                self.ops_async_delay_sec,
            )
            csp.schedule_alarm(alarm, timedelta(seconds=self.ops_async_delay_sec), True)
            s_alarm_scheduled = True
        else:
            log.trace("Heartbeat queue is empty")
            s_alarm_scheduled = False

    @csp.node
    def _publish_alerts(self, data: ts[List[MonitoringEvent]]):
        """ """
        with csp.alarms():
            alarm = csp.alarm(bool)

        with csp.state():
            s_alarm_scheduled = False
            # s_alert queue to store and process async replies from OpsGenie
            s_alert_queue = deque()
            s_alert_api = opsgenie_sdk.AlertApi(api_client=self._api_client)
            s_min_alert_level = MonitoringLevelMapping.from_alert_type(
                level=self.ops_alert_min_level,
            )

        if csp.ticked(data):
            for event in data:
                log.trace("Received MonitoringEvent: %s", event.to_dict())
                event_level = MonitoringLevelMapping.from_event(event)
                # check if the event at or above min reported level
                if event_level.logging >= s_min_alert_level.logging:
                    try:
                        payload = event.to_opsgenie(
                            extra_tags=self.ops_tags,
                            alias_tags=self.ops_alias_tags,
                            category_tag=self.ops_category_tag,
                            separator=self.ops_alias_separator,
                        )
                        log.trace("Processing alert: %s", payload)
                        res = s_alert_api.create_alert(
                            create_alert_payload=payload,
                            async_req=True,
                        )
                        s_alert_queue.append((csp.now(), res))
                        if not s_alarm_scheduled:
                            log.trace(
                                "Setting alarm to check alert queue in %s seconds",
                                self.ops_async_delay_sec,
                            )
                            csp.schedule_alarm(
                                alarm,
                                timedelta(seconds=self.ops_async_delay_sec),
                                True,
                            )
                            s_alarm_scheduled = True
                    except Exception as e:
                        log.error(e, exc_info=True)
                else:
                    log.trace(
                        "Ignoring MonitoringEvent alert level %s (logging level %s) below threshold %s",
                        event.alert_type,
                        event_level,
                        s_min_alert_level,
                    )

        if csp.ticked(alarm):
            # keep processing the first element in the queue until the queue is either empty
            # or the timestamp of the first element is less than ops_async_delay_sec in the past
            while s_alert_queue and (csp.now() - s_alert_queue[0][0]).total_seconds() >= self.ops_async_delay_sec:
                log.trace(
                    "Processing alert queue, %s object(s) remaining",
                    len(s_alert_queue),
                )
                (_, res) = s_alert_queue.popleft()
                if res.ready():
                    log.trace("Alert submission successful")
                else:
                    log.error(
                        "Alert submission is not ready in %s seconds",
                        self.ops_async_delay_sec,
                    )
            log.trace(
                "Done processing alert queue, %s object(s) remaining",
                len(s_alert_queue),
            )
            # if more items remaining reschedule the alarm
            if s_alert_queue:
                log.trace(
                    "Resetting alarm to check alert queue in %s seconds",
                    self.ops_async_delay_sec,
                )
                csp.schedule_alarm(alarm, timedelta(seconds=self.ops_async_delay_sec), True)
                s_alarm_scheduled = True
            else:
                log.trace("Alert queue is empty")
                s_alarm_scheduled = False



from datetime import datetime

from csp import Struct

from csp_gateway import GatewayStruct as Base
from csp_gateway.utils.struct import (
    GatewayLookupMixin,
    GatewayPydanticMixin,
)


class LookupModel(Base):
    foo: int = 9


class NoLookupModel(Base):
    foo: int = 10


NoLookupModel.omit_from_lookup(True)


def test_automatic_id_generation():
    for Model in [LookupModel, NoLookupModel]:
        o1 = Model()
        value1 = str(Model.id_generator.current())
        assert o1.id == value1

        o2 = Model()
        value2 = str(Model.id_generator.current())
        assert o2.id == value2
        assert o2.id == str(int(o1.id) + 1)

        if Model == LookupModel:
            assert Model.lookup(value1) == o1
            assert Model.lookup(value2) == o2


def test_lookup_fails():
    o1 = LookupModel()
    value1 = str(LookupModel.id_generator.current())
    assert o1.id == value1

    o2 = LookupModel()
    value2 = str(LookupModel.id_generator.current())
    assert o2.id == value2
    assert o2.id == str(int(o1.id) + 1)

    assert LookupModel.lookup(value1) == o1
    assert LookupModel.lookup(value2) == o2

    o1 = NoLookupModel()
    value1 = str(NoLookupModel.id_generator.current())
    assert o1.id == value1

    o2 = NoLookupModel()
    value2 = str(NoLookupModel.id_generator.current())
    assert o2.id == value2
    assert o2.id == str(int(o1.id) + 1)

    assert NoLookupModel.lookup(value1) is None
    assert NoLookupModel.lookup(value2) is None


def test_add_lookup_mixin_in_subclass():
    class MyBase(Struct):
        a: int
        id: str
        timestamp: datetime

    # Start with only Pydantic mixin (no lookup or id generator)
    class PydOnly(GatewayPydanticMixin, MyBase):
        pass

    # Provide explicit id/timestamp since no lookup mixin exists to default them
    now = datetime.now()
    p = PydOnly(a=1, id="explicit", timestamp=now)
    # TypeAdapter works without lookup mixin
    p2 = PydOnly.type_adapter().validate_python(p.to_dict())
    assert p2.id == "explicit"
    assert p2.timestamp == now

    # Add lookup mixin later via subclassing
    class WithLookup(GatewayLookupMixin, PydOnly):
        pass

    w = WithLookup(a=2)
    assert isinstance(w.id, str)
    assert isinstance(w.timestamp, datetime)
    assert WithLookup.lookup(w.id) == w
    # generate_id available now
    nid = WithLookup.generate_id()
    assert isinstance(nid, str)


def test_lookup_toggle_isolated_across_inheritance():
    class MyBase(Struct):
        a: int
        id: str
        timestamp: datetime

    class Parent(GatewayLookupMixin, MyBase):
        pass

    # Disable lookup on Parent
    Parent.omit_from_lookup(True)
    p = Parent(a=1)
    assert Parent.lookup(p.id) is None

    # Child inherits mixin; __init_subclass__ should reset include to True
    class Child(Parent):
        pass

    c = Child(a=2)
    assert c.a == 2
    assert Child.lookup(c.id) == c
    # Ensure Parent still disabled
    p2 = Parent(a=3)
    assert p2.a == 3
    assert Parent.lookup(p2.id) is None


def test_lookup_only_mixin_without_fields_mixin():
    class BaseStruct(Struct):
        a: int
        # No fields mixin, declare fields on class
        id: str
        timestamp: datetime

    class LookupOnly(GatewayLookupMixin, BaseStruct):
        pass

    # Defaults applied
    x = LookupOnly(a=5)
    assert isinstance(x.id, str)
    assert isinstance(x.timestamp, datetime)
    assert LookupOnly.lookup(x.id) == x

    # Toggle off lookup
    LookupOnly.omit_from_lookup(True)
    y = LookupOnly(a=6)
    assert LookupOnly.lookup(y.id) is None

    # Toggle back on lookup
    LookupOnly.omit_from_lookup(False)
    z = LookupOnly(a=7)
    assert LookupOnly.lookup(z.id) == z


def test_separate_lookup_registries():
    class StructA(Struct):
        a: int
        id: str
        timestamp: datetime

    class StructB(Struct):
        b: int
        id: str
        timestamp: datetime

    class LookupA(GatewayLookupMixin, StructA):
        pass

    class LookupB(GatewayLookupMixin, StructB):
        pass

    a1 = LookupA(a=1)
    b1 = LookupB(b=1)

    assert LookupA.lookup(a1.id) == a1
    assert LookupB.lookup(b1.id) == b1

    # Cross-lookups must be isolated
    assert LookupA.lookup(b1.id) is None
    assert LookupB.lookup(a1.id) is None

    # Generators are per-class
    a_id1 = LookupA.generate_id()
    a_id2 = LookupA.generate_id()
    b_id1 = LookupB.generate_id()
    b_id2 = LookupB.generate_id()
    assert a_id1 != a_id2
    assert b_id1 != b_id2
    # Not guaranteed, but overwhelmingly likely that sequences don't collide
    assert a_id1 != b_id1
